import sys
import unittest.mock

import usb.core

sys.modules["usb.core"] = unittest.mock.Mock()
sys.modules["usb.core.find"] = unittest.mock.Mock()

from qmi.core.transport_usbtmc_pyusb import QMI_PyUsbTmcTransport
from qmi.core.exceptions import QMI_TimeoutException, QMI_EndOfInputException
# For not causing an error by missing pyvisa library in the pipeline tests, we mock also the pyvisa existence
import tests.core.pyvisa_stub
sys.modules["pyvisa"] = tests.core.pyvisa_stub
sys.modules["pyvisa.errors"] = tests.core.pyvisa_stub.errors
from qmi.core.transport_usbtmc_visa import QMI_VisaUsbTmcTransport
patcher = unittest.mock.patch("pyvisa.errors")


class TestQmiPyUsbTmcTransport(unittest.TestCase):
    """Test QMI_PyUsbTmcTransport."""
    def test_create_object(self):
        dev = QMI_PyUsbTmcTransport(0x1234, 0x5678, "90")
        self.assertIsNotNone(dev)
        self.assertEqual(dev.vendorid, int(0x1234))
        self.assertEqual(dev.productid, int(0x5678))
        self.assertEqual(dev.serialnr, "90")

    @unittest.mock.patch("qmi.core.usbtmc.Instrument")
    def test_open_close(self, mock):
        dev = QMI_PyUsbTmcTransport(0x1234, 0x5678, "90")
        dev.open()
        dev.close()

        mock().open.assert_called_once()
        mock().close.assert_called_once()

    @unittest.mock.patch("qmi.core.usbtmc.Instrument")
    def test_write(self, mock):
        data = b"data"

        dev = QMI_PyUsbTmcTransport(0x1234, 0x5678, "90")
        dev.open()
        dev.write(data)
        dev.close()

        mock().write_raw.assert_called_once_with(data)

    @unittest.mock.patch("qmi.core.usbtmc.Instrument")
    def test_read(self, mock):
        """Test read method works."""
        expected = b"0"
        mock().read_raw.return_value = expected

        dev = QMI_PyUsbTmcTransport(0x1234, 0x5678, "90")
        dev.open()
        ret = dev.read(1, timeout=0.001)
        dev.close()

        mock().read_raw.assert_called_once()
        self.assertEqual(ret, expected)

    @unittest.mock.patch("qmi.core.usbtmc.Instrument")
    def test_read_raises_exception(self, mock):
        """A test to see that read raises QMI_EndOfInputException."""
        # Arrange
        nbytes = 2
        buf = b"0"
        mock().read_raw.return_value = buf

        expected = f"The read message did not contain expected bytes of data ({len(buf)} != {nbytes}."
        dev = QMI_PyUsbTmcTransport(0x1234, 0x5678, "90")
        dev.open()
        with self.assertRaises(QMI_EndOfInputException) as exc:
            dev.read(nbytes, timeout=0.001)

        dev.close()

        mock().read_raw.assert_called_once()
        self.assertEqual(str(exc.exception), str(expected))

    @unittest.mock.patch("qmi.core.usbtmc.Instrument")
    def test_read_until(self, mock):
        """This should forward the executions to read_until_timeout, so we test both in one go."""
        expected = b"01234"

        mock().read_raw.return_value = expected
        dev = QMI_PyUsbTmcTransport(0x1234, 0x5678, "90")
        dev.open()
        ret = dev.read_until(b"T3000", timeout=0.001)
        dev.close()

        mock().read_raw.assert_called_once()
        self.assertEqual(ret, expected)

    @unittest.mock.patch("qmi.core.usbtmc.Instrument")
    def test_discard_read_until_timeout_does_not_raise_timeout_exception(self, mock):
        """This tests that read_until_timeout catches timeout exception and returns empty bytes."""
        usberror = usb.core.USBError("Timeout!")
        usberror.errno = 110  # Set for Timeout
        expected = b""
        mock().read_raw.side_effect = [usberror]

        dev = QMI_PyUsbTmcTransport(0x1234, 0x5678, "90")
        dev.open()
        ret = dev.read_until_timeout(10, timeout=0.001)
        dev.close()

        mock().read_raw.assert_called_once()
        self.assertEqual(ret, expected)

    @unittest.mock.patch("qmi.core.usbtmc.Instrument")
    def test_discard_read(self, mock):
        """See that discard_read tries to read whatever is in the incoming instrument buffer."""
        expected = b"01234"

        mock().read_raw.return_value = expected
        dev = QMI_PyUsbTmcTransport(0x1234, 0x5678, "90")
        dev.open()
        # Discard read should empty the read buffer and discard instrument readout.
        dev.discard_read()
        dev.close()

        mock().read_raw.assert_called_once()

    @unittest.mock.patch("qmi.core.usbtmc.Instrument")
    def test_discard_read_does_not_except(self, mock):
        """See that discard_read passes QMI_TimeoutException if incoming instrument buffer is empty."""
        usberror = usb.core.USBError("Timeout!")
        usberror.errno = 110  # Set for Timeout
        mock().read_raw.side_effect = [usberror]
        dev = QMI_PyUsbTmcTransport(0x1234, 0x5678, "90")
        dev.open()
        dev.discard_read()
        dev.close()

        mock().read_raw.assert_called_once()


class TestQmiVisaUsbTmcTransport(unittest.TestCase):
    """Test QMI_VisaUsbTmcTransport."""
    def test_create_object(self):
        dev = QMI_VisaUsbTmcTransport(0x1234, 0x5678, "90")
        self.assertIsNotNone(dev)

    def test_list_resources(self):
        """Test static method list_resources. Make sure two correct resources are formatted and one
        incorrect resource is ignored."""
        # correct resources
        vendorid_1 = tests.core.pyvisa_stub.vendorid_1
        productid_1 = tests.core.pyvisa_stub.productid_1
        serialnr_1 = tests.core.pyvisa_stub.serialnr_1
        vendorid_3 = tests.core.pyvisa_stub.vendorid_3
        productid_3 = tests.core.pyvisa_stub.productid_3
        serialnr_3 = tests.core.pyvisa_stub.serialnr_3
        # QMI-style transport strings. Resource "2" should be invalid and not included.
        qmi_str_1 = f"usbtmc:vendorid=0x{vendorid_1:04x}:productid=0x{productid_1:04x}:serialnr={serialnr_1}"
        qmi_str_3 = f"usbtmc:vendorid=0x{vendorid_3:04x}:productid=0x{productid_3:04x}:serialnr={serialnr_3}"
        expected_output = [qmi_str_1, qmi_str_3]
        # Act
        output = QMI_VisaUsbTmcTransport.list_resources()
        # Assert
        self.assertListEqual(expected_output, output)

    @unittest.mock.patch("pyvisa.ResourceManager.open_resource")
    def test_open_close(self, mock):
        dev = QMI_VisaUsbTmcTransport(0x1234, 0x5678, "90")
        dev.open()
        dev.close()

        mock.assert_called_once()

    @unittest.mock.patch("pyvisa.ResourceManager.open_resource")
    def test_write(self, mock):
        data = bytes("data".encode())

        dev = QMI_VisaUsbTmcTransport(0x1234, 0x5678, "90")
        dev.open()
        dev.write(data)
        dev.close()

        mock().write_raw.assert_called_once_with(data)

    @unittest.mock.patch("pyvisa.ResourceManager.open_resource")
    def test_read(self, mock):
        expected = b"0"

        mock().read_raw.return_value = expected
        dev = QMI_VisaUsbTmcTransport(0x1234, 0x5678, "90")
        dev.open()
        ret = dev.read(1, timeout=0.001)
        dev.close()

        mock().read_raw.assert_called_once()
        self.assertEqual(ret, expected)

    @unittest.mock.patch("pyvisa.ResourceManager.open_resource")
    @unittest.mock.patch("pyvisa.errors.VisaIOError", new_callable=lambda: tests.core.pyvisa_stub.VisaIOError)
    def test_read_raises_exception(self, exc_mock, mock):
        """A test to see that read raises QMI_TimeoutException """
        # Arrange
        timeout_error = tests.core.pyvisa_stub.VI_ERROR_TMO
        mock().read_raw.side_effect = exc_mock(timeout_error)
        expected = QMI_TimeoutException(str(tests.core.pyvisa_stub.VisaIOError(timeout_error)))
        dev = QMI_VisaUsbTmcTransport(0x1234, 0x5678, "90")
        dev.open()
        with self.assertRaises(QMI_TimeoutException) as exc:
            dev.read(1, timeout=0.001)

        dev.close()

        mock().read_raw.assert_called_once()
        self.assertEqual(repr(exc.exception), repr(expected))

    @unittest.mock.patch("pyvisa.ResourceManager.open_resource")
    def test_read_until(self, mock):
        """This should forward the executions to read_until_timeout, so we test both in one go."""
        expected = b"01234"

        mock().read_raw.return_value = expected
        dev = QMI_VisaUsbTmcTransport(0x1234, 0x5678, "90")
        dev.open()
        ret = dev.read_until(b"T3000", timeout=0.001)
        dev.close()

        mock().read_raw.assert_called_once()
        self.assertEqual(ret, expected)

    @unittest.mock.patch("pyvisa.ResourceManager.open_resource")
    @unittest.mock.patch("pyvisa.errors.VisaIOError", new_callable=lambda: tests.core.pyvisa_stub.VisaIOError)
    def test_read_until_timeout_does_not_raise_timeout_exception(self, exc_mock, mock):
        """This tests that read_until_timeout catches timeout exception and returns empty bytes."""
        timeout_error = tests.core.pyvisa_stub.VI_ERROR_TMO
        expected = b""
        mock().read_raw.side_effect = exc_mock(timeout_error)

        dev = QMI_VisaUsbTmcTransport(0x1234, 0x5678, "90")
        dev.open()
        ret = dev.read_until_timeout(10, timeout=0.001)
        dev.close()

        mock().read_raw.assert_called_once()
        self.assertEqual(ret, expected)

    @unittest.mock.patch("pyvisa.ResourceManager.open_resource")
    def test_discard_read(self, mock):
        """See that discard_read empties the read buffer, and
        it tries to read whatever is in the incoming instrument buffer.
        """
        expected = b"01234"

        mock().read_raw.return_value = expected
        dev = QMI_VisaUsbTmcTransport(0x1234, 0x5678, "90")
        dev.open()
        # Discard read should empty the read buffer and read instrument data
        dev.discard_read()
        dev.close()

        mock().read_raw.assert_called_once()

    @unittest.mock.patch("pyvisa.ResourceManager.open_resource")
    @unittest.mock.patch("pyvisa.errors.VisaIOError", new_callable=lambda: tests.core.pyvisa_stub.VisaIOError)
    def test_discard_read_does_not_except(self, exc_mock, mock):
        """See that discard_read passes QMI_TimeoutException if read buffer and incoming instrument buffer are empty."""
        timeout_error = tests.core.pyvisa_stub.VI_ERROR_TMO
        mock().read_raw.side_effect = exc_mock(timeout_error)
        dev = QMI_VisaUsbTmcTransport(0x1234, 0x5678, "90")
        dev.open()
        dev.discard_read()
        dev.close()

        mock().read_raw.assert_called_once()


if __name__ == "__main__":
    unittest.main()



"""
Instrument driver for the PicoTech PicoScope 4000A series oscilloscopes.

This driver depends on the PicoSDK libraries from PicoTech and on
the PicoSDK Python wrappers from
https://github.com/picotech/picosdk-python-wrappers .
"""

import ctypes
import logging
from typing import TYPE_CHECKING

from qmi.core.context import QMI_Context
from qmi.core.rpc import rpc_method

from qmi.instruments.picotech._picoscope import PicoTech_PicoScope, _check_error

# Lazy import of the "picosdk" module. See the function _import_modules() in _picoscope.py.
if TYPE_CHECKING:
    from picosdk.ps4000a import ps4000a as _ps
else:
    _ps = None


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class PicoTech_PicoScope4824(PicoTech_PicoScope):
    """Instrument driver for the PicoTech PicoScope 4824 USB oscilloscope.

    Attributes:
        NUM_CHANNELS:     Number of oscilloscope channels.
        NUM_INPUT_RANGES: Number of supported input ranges. Range '0' is not supported.
    """
    NUM_CHANNELS = 8
    NUM_INPUT_RANGES = 12

    def __init__(self, context: QMI_Context, name: str, serial_number: str) -> None:
        """Initialize the instrument driver.

        Arguments:
            name: Name for this instrument instance.
            serial_number: Serial number of the Picoscope.
        """
        super().__init__(context, name, serial_number)
        self._library = "4000a"

    @rpc_method
    def run_block(self, num_pretrig_samples: int, num_posttrig_samples: int, time_base: int) -> None:
        """Start acquisition in block mode.

        If the trigger is enabled (see `set_trigger()`), the oscilloscope will
        until the trigger condition occurs, then acquire a single block of data.
        If the trigger is disabled, the oscilloscope will immediately acquire
        a single block of data.

        This function returns immediately while the acquisition is in progress.
        Call `is_block_ready()` or `wait_block_ready()` to check whether the acquisition has finished.
        Optionally call `stop()` to end the acquisition before it completes.

        Parameters:
            num_pretrig_samples:    Number of pre-trigger samples.
            num_posttrig_samples:   Number of post-trigger samples.
            time_base:              Timebase selector (range 0 ... 2**32-1).
                                    The effective timebase is `(timebase+1) * 12.5 ns`.
                                    Depending on the enabled channels, a minimum value of 1 may be required
                                    (minimum timebase 25 ns).
        """
        self._check_is_open()

        num_samples = num_pretrig_samples + num_posttrig_samples
        par_interval_ns = ctypes.c_float()
        err = _ps.ps4000aGetTimebase2(
            self._handle,
            time_base,
            num_samples,
            ctypes.byref(par_interval_ns),
            None,  # maxSamples
            0,  # segmentIndex
        )
        _check_error(err)

        err = _ps.ps4000aRunBlock(
            self._handle,
            num_pretrig_samples,
            num_posttrig_samples,
            time_base,
            None,  # timeIndisposedMs
            0,     # segmentIndex
            None,  # lpReady
            None,  # pParameter
        )
        _check_error(err)

        self._num_samples = num_samples
        self._timebase_interval_ns = par_interval_ns.value

    @rpc_method
    def get_sampling_interval(self, time_base: int) -> float:
        """Returns the scope's time resolution in nanoseconds depending on the time_base selector.

        Parameters:
            time_base:   Timebase selector (range 0 ... 2^32-1).
                         The effective time resolution is 12.5 ns * (time_base + 1).

        Returns:
            resolution:  Scope's time resolution in nanoseconds.
                         0 is returned for invalid time_base.
        """
        if 0 <= time_base < 2 ** 32:
            return 12.5 * (time_base + 1)

        else:
            return 0



"""
Anapico, signal generators.

The qmi.instruments.agiltron package provides support for:
- APSIN Series signal generators.
"""
# Alternative, QMI naming convention approved name
from qmi.instruments.anapico.apsin import Anapico_APSIN as Anapico_Apsin



"""
Agiltron, optical switch driver device.

The qmi.instruments.agiltron package provides support for:
- FF 1x4 optical switch driver
"""
# Using QMI naming convention approved name
from qmi.instruments.agiltron.ff1x4 import Agiltron_FF1x4 as Agiltron_Ff1x4



#! /usr/bin/env python3

"""Unit test for T3 mode event processing in HydraHarp driver."""

import ctypes
import threading
import unittest
from unittest.mock import patch, PropertyMock

import numpy as np

from qmi.core.exceptions import QMI_RuntimeException
from qmi.core.pubsub import QMI_SignalReceiver
from qmi.instruments.picoquant.support._hhlib_function_signatures import _hhlib_function_signatures
from qmi.instruments.picoquant import EventDataType, EventFilterMode
from qmi.instruments.picoquant import PicoQuant_HydraHarp400

from tests.patcher import PatcherQmiContext as QMI_Context


def gen_events(max_events, sync_period):
    events_in = np.empty(max_events, dtype=EventDataType)
    events_in["type"] = 0
    events_in["timestamp"] = np.cumsum(np.random.randint(1, sync_period, max_events))
    # At each passing of a sync moment we need to add a sync signal
    (inds,) = np.where((
            events_in["timestamp"] // sync_period > np.insert(events_in["timestamp"], 0, 0)[:-1] // sync_period
    ))
    # And then insert the sync signals and cap the length
    events_in["type"] = np.insert(events_in["type"], inds - 1, np.array([64] * len(inds)))[:max_events]
    events_in["timestamp"] = np.insert(
        events_in["timestamp"], inds - 1,
                                (events_in["timestamp"] // sync_period)[inds - 1] * sync_period
    )[:max_events]
    events_in = events_in[np.lexsort((-1 * events_in["type"].astype("int16"), events_in["timestamp"]))]
    # This can leave the last index wrong. Check.
    last_sync_index = -(events_in["type"][::-1].tolist().index(64) + 1)
    if (events_in["timestamp"][-1] - events_in["timestamp"][last_sync_index]) > sync_period:
        # we replace the last input with a sync
        events_in["type"][-1] = 64
        events_in["timestamp"][-1] = events_in["timestamp"][last_sync_index] + sync_period
        events_in = events_in[np.lexsort((-1 * events_in["type"].astype("int16"), events_in["timestamp"]))]

    return events_in


def gen_random_events(num_events, event_rate, sync_rate, resolution, channels, seed=None):
    rng = np.random.default_rng(seed)

    mean_event_interval = event_rate // num_events // resolution
    delta_times = rng.exponential(scale=mean_event_interval, size=num_events)
    delta_times = delta_times.astype(np.uint64)
    event_timestamps = np.cumsum(delta_times)

    sync_period = 1e12 / sync_rate
    syncs_in_events = int(event_timestamps.max() // sync_period)
    sync_timestamps = np.linspace(0, sync_period * syncs_in_events, num=(syncs_in_events + 1))
    # Filter out only those syncs that have an event within sync range
    inds = [i for i in range(0, len(sync_timestamps) - 2) if np.any(
        (event_timestamps > sync_timestamps[i]) & (sync_timestamps[i + 1] > event_timestamps)
    )]
    sync_timestamps = sync_timestamps[np.array(inds)]
    sync_timestamps = sync_timestamps.astype(np.uint64)

    events = np.empty(num_events + len(sync_timestamps), dtype=EventDataType)
    events[:num_events]["type"] = rng.choice(channels, size=num_events)
    events[:num_events]["timestamp"] = event_timestamps
    events[num_events:]["type"] = 64
    events[num_events:]["timestamp"] = sync_timestamps
    events = np.unique(events)  # Remove any duplicates
    events = events[np.lexsort((-1 * events["type"].astype("int16"), events["timestamp"]))]

    return events[:num_events]


def events_to_fifo(events, sync_period, resolution):
    """Convert array of events to the corresponding HydraHarp T3 FIFO data words."""

    # Convert events to array (if necessary).
    evt_array = np.array([_ for _ in events if _[0] != 64], dtype=EventDataType)

    # Check monotonic timestamps.
    assert np.all(evt_array["timestamp"][1:] >= evt_array["timestamp"][:-1])

    tmp_types = evt_array["type"]
    tmp_timestamps = evt_array["timestamp"]

    # count "timestamp" residuals over sync moments for nsync.
    nsync_counts = (tmp_timestamps // sync_period - np.insert(tmp_timestamps, 0, 0)[:-1] // sync_period)
    # In bits 10:25 will be added the dtime.
    dtime_counts = (tmp_timestamps % sync_period).astype(np.uint64) // int(resolution)
    # Filter out timestamps outside sync period range.
    relative_timestamp = tmp_timestamps - np.cumsum(nsync_counts) * sync_period
    valid_ndx = relative_timestamp < sync_period
    # We have to remove any entries that are larger than the 2^15 - 1 (32767 * R) range
    valid_ndx = valid_ndx & (dtime_counts < (2 ** 15 - 1) / resolution)
    # Filter out invalid indexes for d_time and also for time stamps.
    nsync_counts = nsync_counts[valid_ndx]
    dtime_counts = dtime_counts[valid_ndx]
    tmp_timestamps = tmp_timestamps[valid_ndx]

    # Insert overflow records at positions where the MSB part of the timestamp changes.
    overflows = (
            tmp_timestamps // sync_period // 1024
            - np.insert(tmp_timestamps, 0, 0)[:-1] // sync_period // 1024
    )
    (overflow_idx,) = np.where(overflows > 0)
    overflow_counts = overflows[overflow_idx]
    # Have to set counts, special bit (31 from this side) for overflow, and the channel number (63 for overflow)
    overflow_words = overflow_counts.astype(np.uint64) & 0x01ffffff | ((1 << 31) | (63 << 25))
    # Convert events to FIFO words. Include "type" for channel number.  This has to be all done in one go.
    fifo_words = (
            (tmp_types[valid_ndx].astype(np.uint64) << 25) |  # Shift the channel numbers to bits 25:30
            (dtime_counts.astype(np.uint64) << 10) |  # dtimes at bits 10:24
            np.cumsum(nsync_counts).astype(np.uint64) & 0x03ff  # nsync at bits 0:9
    )
    # And then insert the overflow_words into it:
    fifo_words = np.insert(fifo_words, overflow_idx, overflow_words)

    return fifo_words


def make_patched_read_fifo(fifo_words, done_event):
    """Return a function to be used in place of the actual readFifo() method in the HydraHarp API."""

    fifo_pos = [0]

    def patched_read_fifo(_, fifo_data, ttreadmax, nactual):
        pos = fifo_pos[0]
        n_fifo_words = len(fifo_words)
        npos = min(n_fifo_words, pos + ttreadmax)  # read 131072 words. Resets timestamp after npos words.
        fifo_pos[0] = npos
        for i, val in enumerate(fifo_words[pos:npos]):
            fifo_data[i] = val
        nactual.value = npos - pos
        if (done_event is not None) and (pos == n_fifo_words):
            done_event.set()
        return 0

    return patched_read_fifo


class TestHydraHarpEventFilter(unittest.TestCase):

    # Reduced limit on maximum pending events to make it easier to test.
    PATCHED_MAX_PENDING_EVENTS = 1024
    PATCHED_MAX_EVENTS_PER_CALL = 100000  # Should be less than TTREADMAX
    SYNC_FREQUENCY_32MHz = 32e6  # Use 32MHz sync frequency.
    SYNC_FREQUENCY_5MHz = 5e6  # Use 5MHz sync frequency.
    RESOLUTION_LIMIT = 2**15 - 1   # Limit for 32768*R resolution limit

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Generate random event sets to be used in test cases.
        self.events_in_resolution_1 = gen_random_events(
            num_events=1000,
            event_rate=2.0e6,
            sync_rate=self.SYNC_FREQUENCY_32MHz,  # creates a sync period of 31250 < 2**15 - 1
            resolution=1,
            channels=[0, 1],
            seed=2345,
        )

        self.events_in_4chan_resolution_1 = gen_random_events(
            num_events=50000,
            event_rate=1.0e9,
            sync_rate=self.SYNC_FREQUENCY_32MHz,
            resolution=1,
            channels=[0, 1, 2, 3],
            seed=1234,
        )

    def setUp(self) -> None:
        patcher = patch('qmi.instruments.picoquant.PicoQuant_HydraHarp400._lib', new_callable=PropertyMock)
        self._library_mock = patcher.start().return_value
        self.addCleanup(patcher.stop)

        function_names, _, _ = zip(*_hhlib_function_signatures)
        stripped_names = [name.split("_")[1] for name in function_names]
        self._library_mock.mock_add_spec(stripped_names, spec_set=True)

        # Create QMI context.
        self._ctx = QMI_Context("hydraharp_eventfilter_test")

        # Create instrument instance. Limit MAX calls to avoid problems.
        PicoQuant_HydraHarp400.MAX_EVENTS_PER_CALL = self.PATCHED_MAX_EVENTS_PER_CALL
        if self._testMethodName == "test_pending_events_limit":
            self._hydraharp: PicoQuant_HydraHarp400 = PicoQuant_HydraHarp400(
                self._ctx,
                "hydraharp",
                "1111111",
                max_pending_events=self.PATCHED_MAX_PENDING_EVENTS
            )
        else:
            self._hydraharp: PicoQuant_HydraHarp400 = PicoQuant_HydraHarp400(
                self._ctx,
                "hydraharp",
                "1111111"
            )

        self._library_mock.GetLibraryVersion.return_value = 0
        self._library_mock.OpenDevice.return_value = 0

        string_buffer = ctypes.create_string_buffer(b"1111111")
        with patch("sys.platform", "linux1"), patch(
            "ctypes.create_string_buffer", return_value=string_buffer
        ):
            self._hydraharp.open()

        self._hydraharp.initialize("T3", "EXTERNAL")

    def tearDown(self):
        # Close instrument instance.
        self._hydraharp.close()
        self._ctx = None

    def test_simple(self):
        """Simple test with default fixed sync frequency and resolution being 1ps."""
        # Events to use during this test. In T3 mode there is in principle one sync event per channel.
        # There are thus 1 sync event per sync period, which is 1/self.SYNC_FREQUENCY * 1E12 (2E5 ps at 5MHz)
        sync_period = 1e12 / self.SYNC_FREQUENCY_5MHz
        events_in = [
            (64, 0),
            (0, int(self.RESOLUTION_LIMIT // 3)),
            (1, int(self.RESOLUTION_LIMIT // 2)),
            (64, int(sync_period * 5000)),
            (0, int(sync_period * 5000 + self.RESOLUTION_LIMIT // 3)),
            (0, int(sync_period * 5000 + self.RESOLUTION_LIMIT // 2)),
            (64, int(sync_period * 6025)),
            (1, int(sync_period * 6025 + self.RESOLUTION_LIMIT // 4)),
            (1, int(sync_period * 6025 + self.RESOLUTION_LIMIT // 2)),
        ]

        # Patch the readFifo() function.
        fifo_words_in = events_to_fifo(events_in, sync_period, 1.0)  # resolution = 1
        done_event = threading.Event()
        self._library_mock.ReadFiFo.side_effect = make_patched_read_fifo(fifo_words_in, done_event)

        # Start the measurement.
        sync_rate = ctypes.c_int(int(self.SYNC_FREQUENCY_5MHz))
        resolution = ctypes.c_double(1.0)
        with patch("ctypes.c_double", return_value=resolution), patch(
            "ctypes.c_int", return_value=sync_rate
        ):
            self._hydraharp.start_measurement(1000)

        # Wait to make sure all events are processed.
        done_event.wait()

        # Stop the measurement.
        self._hydraharp.stop_measurement()

        # Fetch events.
        events = self._hydraharp.get_events()

        # Check events.
        self.assertIsInstance(events, np.ndarray)
        self.assertEqual(events.dtype, EventDataType)
        self.assertTrue(np.all(events == np.array(events_in, dtype=EventDataType)))

    def test_resolution_simple(self):
        """Simple test with default fixed sync frequency and resolution being 1ps."""
        # Events to use during this test. In T3 mode there is in principle one sync event per channel.
        # There are thus 1 sync event per sync period, which is 1/self.SYNC_FREQUENCY * 1E12 (2E5 ps at 5MHz)
        sync_rate = ctypes.c_int(int(self.SYNC_FREQUENCY_5MHz))
        sync_period = 1e12 / sync_rate.value
        resolution = ctypes.c_double(2.0)
        events_in = [
            (64, 0),
            (0, int(self.RESOLUTION_LIMIT // 4)),
            (1, int(self.RESOLUTION_LIMIT // 3)),
            (64, int(sync_period * 5000)),
            (0, int(sync_period * 5000 + self.RESOLUTION_LIMIT // 4)),
            (0, int(sync_period * 5000 + self.RESOLUTION_LIMIT // 3)),
            (64, int(sync_period * 6025)),
            (1, int(sync_period * 6025 + self.RESOLUTION_LIMIT // 5)),
            (1, int(sync_period * 6025 + self.RESOLUTION_LIMIT // 3)),
        ]
        expected_events = np.array(events_in, dtype=EventDataType)
        for e in (1, 4, 7):
            expected_events[e][1] -= 1

        # Patch the readFifo() function.
        fifo_words_in = events_to_fifo(events_in, sync_period, resolution.value)
        done_event = threading.Event()
        self._library_mock.ReadFiFo.side_effect = make_patched_read_fifo(fifo_words_in, done_event)

        # Start the measurement.
        with patch("ctypes.c_double", return_value=resolution), patch(
            "ctypes.c_int", return_value=sync_rate
        ):
            self._hydraharp.start_measurement(1000)

        # Wait to make sure all events are processed.
        done_event.wait()

        # Stop the measurement.
        self._hydraharp.stop_measurement()

        # Fetch events.
        events = self._hydraharp.get_events()

        # Check events.
        self.assertIsInstance(events, np.ndarray)
        self.assertEqual(events.dtype, EventDataType)
        self.assertTrue(np.all(events == expected_events))

    def test_simple_clip_resolution(self):
        """Simple test with default fixed sync frequency and resolution being 4ps.
        Most of the sample events need to be modified to be a multiple of 4 to get exact same numbers out.
        """
        # Events to use during this test. In T3 mode there is in principle one sync event per channel.
        # There are thus 1 sync event per sync period, which is 1/self.SYNC_FREQUENCY * 1E12 (2E5 ps at 5MHz)
        sync_rate = ctypes.c_int(int(self.SYNC_FREQUENCY_5MHz))
        sync_period = 1e12 / sync_rate.value
        resolution = ctypes.c_double(4.0)
        events_in = [
            (64, 0),
            (1, int(self.RESOLUTION_LIMIT // 3 + 2)),
            (0, int(self.RESOLUTION_LIMIT // 2 + 1)),
            (1, int(self.RESOLUTION_LIMIT + 1)),  # should be clipped
            (64, int(sync_period * 5000)),
            (0,  int(sync_period * 5000 + self.RESOLUTION_LIMIT // 3 + 2)),
            (0,  int(sync_period * 5000 + self.RESOLUTION_LIMIT // 2 + 1)),
            (0,  int(sync_period * 5000 + self.RESOLUTION_LIMIT + 1)),  # should be clipped
            (64, int(sync_period * 6025)),
            (0,  int(sync_period * 6025 + self.RESOLUTION_LIMIT // 4 + 1)),
            (1,  int(sync_period * 6025 + self.RESOLUTION_LIMIT // 2 + 1)),
            (1,  int(sync_period * 6025 + self.RESOLUTION_LIMIT))  # Cannot add, edit later
        ]
        expected_events = np.delete(np.array(events_in, dtype=EventDataType), [3, 7])
        expected_events[-1][1] -= 3  # rounds down to closest multiple of 4
        # Patch the readFifo() function.
        fifo_words_in = events_to_fifo(events_in, sync_period, resolution.value)  # resolution = 1
        done_event = threading.Event()
        self._library_mock.ReadFiFo.side_effect = make_patched_read_fifo(fifo_words_in, done_event)

        # Start the measurement.
        with patch("ctypes.c_double", return_value=resolution), patch(
            "ctypes.c_int", return_value=sync_rate
        ):
            self._hydraharp.start_measurement(1000)

        # Wait to make sure all events are processed.
        done_event.wait()

        # Stop the measurement.
        self._hydraharp.stop_measurement()

        # Fetch events.
        events = self._hydraharp.get_events()

        # Check events.
        self.assertIsInstance(events, np.ndarray)
        self.assertEqual(events.dtype, EventDataType)
        self.assertTrue(np.all(events == expected_events))

    def test_channel_filter(self):
        """Test filtering by setting channels 1 and 2 to be filtered out."""
        sync_rate = ctypes.c_int(int(self.SYNC_FREQUENCY_32MHz))
        resolution = ctypes.c_double(1.0)
        sync_period = 1e12 / sync_rate.value
        # Generate a random set of events.
        events_in = self.events_in_4chan_resolution_1

        # Patch the readFifo() function.
        fifo_words_in = events_to_fifo(events_in, sync_period, resolution.value)
        done_event = threading.Event()
        self._library_mock.ReadFiFo.side_effect = make_patched_read_fifo(fifo_words_in, done_event)

        # Configure event filter: block events form channels 1 and 2.
        self._hydraharp.set_event_filter(
            channel_filter={1: EventFilterMode.NO_EVENTS, 2: EventFilterMode.NO_EVENTS}
        )

        # Start the measurement.
        with patch("ctypes.c_double", return_value=resolution), patch(
            "ctypes.c_int", return_value=sync_rate
        ):
            self._hydraharp.start_measurement(1000)

        # Wait to make sure all events are processed.
        done_event.wait()

        # Stop the measurement.
        self._hydraharp.stop_measurement()

        # Fetch events.
        events = self._hydraharp.get_events()

        # Expect only events on SYNC and channels 0 and 1. Also sort.
        events_expected = events_in[
            (events_in["type"] == 64)
            | (events_in["type"] == 0)
            | (events_in["type"] == 3)
        ]
        events_expected = events_expected[
            np.lexsort((-1 * events_expected["type"].astype("int16"), events_expected["timestamp"]))
        ]
        # Check events.
        self.assertTrue(np.all(events == events_expected))

    def test_aperture_filter_events(self):
        """Filter out events that do not fall withing a specific 'aperture' time per channel."""
        sync_rate = ctypes.c_int(int(self.SYNC_FREQUENCY_32MHz))
        resolution = ctypes.c_double(1.0)
        sync_period = 1e12 / sync_rate.value
        num_events = 50000
        event_rate = 2.0e9
        # Generate a random set of events.
        events_in = gen_random_events(
            num_events=num_events,
            event_rate=event_rate,
            sync_rate=sync_rate.value,
            resolution=resolution.value,
            channels=[0, 1],
            seed=2345,
        )

        # Patch the readFifo() function.
        fifo_words_in = events_to_fifo(events_in, sync_period, resolution.value)
        done_event = threading.Event()
        self._library_mock.ReadFiFo.side_effect = make_patched_read_fifo(fifo_words_in, done_event)

        # Configure event filter: setup aperture filtering for channels 0 and 1.
        delta_min = int(25000 // resolution.value)
        delta_max = int(75000 // resolution.value)
        self._hydraharp.set_event_filter(
            channel_filter={0: EventFilterMode.APERTURE, 1: EventFilterMode.APERTURE},
            sync_aperture=(delta_min, delta_max),
        )

        # Start the measurement.
        with patch("ctypes.c_double", return_value=resolution), patch(
            "ctypes.c_int", return_value=sync_rate
        ):
            self._hydraharp.start_measurement(1000)

        # Wait to make sure all events are processed.
        done_event.wait()

        # Stop the measurement.
        self._hydraharp.stop_measurement()

        # Fetch events.
        events = self._hydraharp.get_events()

        # Expect only events on SYNC and events that fall in the aperture.
        events_expected = []
        last_sync = -1
        for evt_type, evt_timestamp in events_in:
            if evt_type == 64:
                last_sync = evt_timestamp
                events_expected.append((evt_type, evt_timestamp))
            else:
                if (last_sync >= 0) and (delta_min <= (evt_timestamp - last_sync) <= delta_max):
                    events_expected.append((evt_type, evt_timestamp))

        # Check events.
        self.assertTrue(np.all(events == np.array(events_expected, dtype=EventDataType)))

    def test_aperture_filter_sync(self):
        sync_rate = ctypes.c_int(int(self.SYNC_FREQUENCY_32MHz))
        resolution = ctypes.c_double(1.0)
        sync_period = 1e12 / sync_rate.value
        # Generate a random set of events.
        events_in = self.events_in_resolution_1

        # Patch the readFifo() function.
        fifo_words_in = events_to_fifo(events_in, sync_period, resolution.value)
        done_event = threading.Event()
        self._library_mock.ReadFiFo.side_effect = make_patched_read_fifo(fifo_words_in, done_event)

        # Configure event filter: setup aperture filtering for channels 0 and 1 and SYNC.
        delta_min = 25000
        delta_max = 75000
        self._hydraharp.set_event_filter(
            channel_filter={
                0: EventFilterMode.APERTURE,
                1: EventFilterMode.APERTURE,
                64: EventFilterMode.APERTURE,
            },
            sync_aperture=(delta_min, delta_max),
        )

        # Start the measurement.
        with patch("ctypes.c_double", return_value=resolution), patch(
            "ctypes.c_int", return_value=sync_rate
        ):
            self._hydraharp.start_measurement(1000)

        # Wait to make sure all events are processed.
        done_event.wait()

        # Stop the measurement.
        self._hydraharp.stop_measurement()

        # Fetch events.
        events = self._hydraharp.get_events()

        # Expect only events on SYNC and events that fall in the aperture.
        events_expected = []
        last_sync = -1
        last_is_sync = False
        for evt_type, evt_timestamp in events_in:
            if evt_type == 64:
                if last_is_sync:
                    events_expected.pop()

                last_sync = evt_timestamp
                last_is_sync = True
                events_expected.append((evt_type, evt_timestamp))

            else:
                if (last_sync >= 0) and (delta_min <= evt_timestamp - last_sync <= delta_max):
                    events_expected.append((evt_type, evt_timestamp))
                    last_is_sync = False

        if last_is_sync:
            events_expected.pop()

        # Check events.
        self.assertTrue(np.all(events == np.array(events_expected, dtype=EventDataType)))

    def test_get_events_limit(self):
        """Test that the `get_events()` function returns at most MAX_EVENTS_PER_CALL events."""
        sync_rate = ctypes.c_int(int(self.SYNC_FREQUENCY_32MHz))
        resolution = ctypes.c_double(4.0)
        sync_period = 1e12 / sync_rate.value
        # Generate random events.
        final_block_events = 10101
        events_per_call = PicoQuant_HydraHarp400.MAX_EVENTS_PER_CALL
        num_events = 2 * events_per_call + final_block_events
        events_in = np.empty(num_events, dtype=EventDataType)
        events_in["type"] = 0
        events_in["timestamp"] = np.cumsum(np.random.randint(1, 1000, num_events) * int(resolution.value))
        # The events will be passed to the driver in smaller series, in blocks of TTREADMAX.
        # Calculate the sync moments and add into the expected events that should come out
        expected_events = np.array(events_in, dtype=EventDataType)
        sync_counts = (
            expected_events["timestamp"] // sync_period
            - np.insert(expected_events["timestamp"], 0, 0)[:-1] // sync_period
        )
        sync_idx = np.where(sync_counts > 0)[0]
        # The above does not take into account the first possible sync timestamp at 0. Check.
        if sync_idx[0] > 0 and expected_events["timestamp"][0] < sync_period:
            sync_idx = np.insert(sync_idx, 0, 0)

        cum_counts = np.cumsum(sync_counts, dtype=np.uint64)
        # Problem: with resolution of 4, every second sync moment falls on "2".
        # So, every second block of timestamp data is shifted by 2. We need to take this into account.
        timestamp_shift = 2 * (cum_counts % 2)
        expected_events["timestamp"] -= timestamp_shift

        sync_events = np.empty(len(sync_idx), dtype=EventDataType)
        sync_events["type"] = 64
        sync_events["timestamp"] = cum_counts[sync_idx] * sync_period
        expected_events = np.insert(expected_events, sync_idx, sync_events)
        expected_events = expected_events[
            np.lexsort((-1 * expected_events["type"].astype("int16"), expected_events["timestamp"]))
        ]

        # Patch the readFifo() function.
        fifo_words_in = events_to_fifo(events_in, sync_period, resolution.value)
        done_event = threading.Event()
        self._library_mock.ReadFiFo.side_effect = make_patched_read_fifo(fifo_words_in, done_event)

        # Start the measurement.
        with patch("ctypes.c_double", return_value=resolution), patch(
            "ctypes.c_int", return_value=sync_rate
        ):
            self._hydraharp.start_measurement(1000)

        # Wait to make sure all events are processed.
        done_event.wait()
        done_event.clear()

        # Stop the measurement.
        self._hydraharp.stop_measurement()

        # Fetch events.
        # It should return only the first MAX_EVENTS_PER_CALL events.
        events = self._hydraharp.get_events()

        # Check events.
        self.assertIsInstance(events, np.ndarray)
        self.assertEqual(events.dtype, EventDataType)
        self.assertEqual(len(events), events_per_call)
        self.assertTrue(np.all(events == expected_events[:events_per_call]))

        # Now get the next MAX_EVENTS_PER_CALL events.
        (_ts, events) = self._hydraharp.get_timestamped_events()
        # Find and remove any duplicates (possibly a bug of the test, but also possibly of the event fetching thread)
        vals, inverse, count = np.unique(events, return_inverse=True, return_counts=True)
        idx_vals_repeated = np.where(count > 1)[0]
        if idx_vals_repeated.shape[0] > 0:
            # duplicates[0] is an enumeration of duplicate values found and [1] the indices of the duplicates
            duplicates = np.where(inverse == idx_vals_repeated[:, np.newaxis])
            # As the enum repeats the same number as many times as value is found, and we want also to make sure we take
            # into account also the cases where there are more than one duplicate, we find with np.where how many same
            # values are present; with `+ 1` to shift the indices to remove other than the 1st (i.e. the original first)
            # value's duplicate(s).
            deletion_array = duplicates[1][np.where((duplicates[0][1:] - duplicates[0][:-1]) == 0)[0] + 1]
            events = np.delete(events, deletion_array)

        len_events = len(events)
        self.assertAlmostEqual(len_events, events_per_call, delta=1)  # delta takes into account that possibly one entry
        delta = events_per_call - len_events  # is removed.
        self.assertTrue(np.all(events == expected_events[events_per_call:2*events_per_call-delta]))

        # Get the final batch of events. Reduce the added sync events from total count
        events = self._hydraharp.get_events()
        self.assertEqual(len(events) - np.count_nonzero(expected_events["type"] == 64), final_block_events + 1)
        self.assertTrue(np.all(events == expected_events[2*events_per_call-delta:]))

        # Check no further events.
        events = self._hydraharp.get_events()
        self.assertIsInstance(events, np.ndarray)
        self.assertEqual(len(events), 0)

    def test_pending_events_limit(self):
        """Test that the driver correctly handles overflow of the pending event buffer. Resolution is 1"""
        sync_period = int(1e12 // self.SYNC_FREQUENCY_32MHz)
        # Generate a bunch of events that just barely fit into the pending event buffer.
        max_events = self.PATCHED_MAX_PENDING_EVENTS - 1
        events_in = gen_events(max_events, sync_period)

        # Patch the readFifo() function.
        # This fill feed the first part of the events to the driver (not enough to overflow the buffer).
        resolution = ctypes.c_double(1.0)
        fifo_words_in = events_to_fifo(events_in, sync_period, resolution.value)
        done_event = threading.Event()
        self._library_mock.ReadFiFo.side_effect = make_patched_read_fifo(fifo_words_in, done_event)

        # Start the measurement.
        sync_rate = ctypes.c_int(int(self.SYNC_FREQUENCY_32MHz))
        with patch("ctypes.c_double", return_value=resolution), patch(
            "ctypes.c_int", return_value=sync_rate
        ):
            self._hydraharp.start_measurement(1000)

        # Wait to make sure all queued events are processed.
        done_event.wait()
        done_event.clear()

        # Fetch the events queued so far.
        events = self._hydraharp.get_events()
        # Check events.
        # if len(events) < len(events_in):
        # The `get_events` cannot give a sync signal as the last entry. If that is the case for events_in, remove it
        if events_in["type"][-1] == 64:
            events_in = events_in[:-1]

        self.assertEqual(len(events), len(events_in))
        self.assertTrue(np.all(events == events_in))

        # Generate more events than can fit into the pending event buffer.
        more_events = self.PATCHED_MAX_PENDING_EVENTS + 2
        more_events_in = gen_events(more_events, sync_period)

        # Repatch the readFifo() function to feed the new events into the running measurement.
        # This should/will cause the pending event buffer to overflow.
        more_fifo_words_in = events_to_fifo(more_events_in, sync_period, resolution.value)
        self._library_mock.ReadFiFo.side_effect = make_patched_read_fifo(more_fifo_words_in, done_event)
        # Wait to make sure all queued events are processed.
        done_event.wait()

        # Try to fetch events. This should raise an exception as a result of the overflow.
        with self.assertRaises(QMI_RuntimeException):
            _ = self._hydraharp.get_events()

        # Stop the measurement.
        done_event.clear()
        self._hydraharp.stop_measurement()

        # Generate new events to check that a subsequent measurement runs cleanly after overflow.
        events_in = gen_events(max_events, sync_period)

        fifo_words_in = events_to_fifo(events_in, sync_period, resolution.value)
        self._library_mock.ReadFiFo.side_effect = make_patched_read_fifo(fifo_words_in, done_event)

        # Start the measurement.
        sync_rate = ctypes.c_int(int(self.SYNC_FREQUENCY_32MHz))  # Redefine this as otherwise it gets corrupted
        with patch("ctypes.c_double", return_value=resolution), patch("ctypes.c_int", return_value=sync_rate):
            self._hydraharp.start_measurement(1000)

        # Wait to make sure all queued events are processed.
        done_event.wait()

        # Fetch the events queued so far.
        events = self._hydraharp.get_events()

        # Check events.
        # if len(events) < len(events_in):
        # The `get_events` cannot give a sync signal as the last entry. If that is the case for events_in, remove it
        if events_in["type"][-1] == 64:
            events_in = events_in[:-1]

        self.assertEqual(len(events), len(events_in))
        self.assertTrue(np.all(events == events_in))

        # Stop the measurement.
        self._hydraharp.stop_measurement()

        # Check no further events.
        events = self._hydraharp.get_events()
        self.assertEqual(len(events), 0)


class TestHydraHarpRealtime(unittest.TestCase):
    """Test class to test real-time functionalities of event processing. This requires a running QMI context.
    """
    # Reduced limit on maximum pending events to make it easier to test.
    PATCHED_MAX_PENDING_EVENTS = 1024
    PATCHED_MAX_EVENTS_PER_CALL = 100000  # Should be less than TTREADMAX
    SYNC_FREQUENCY_32MHz = 32e6  # Use 32MHz sync frequency.
    SYNC_FREQUENCY_5MHz = 5e6  # Use 5MHz sync frequency.
    RESOLUTION_LIMIT = 2 ** 15 - 1  # Limit for 32768*R resolution limit

    def setUp(self) -> None:
        patcher = patch('qmi.instruments.picoquant.PicoQuant_HydraHarp400._lib', new_callable=PropertyMock)
        self._library_mock = patcher.start().return_value
        self.addCleanup(patcher.stop)

        function_names, _, _ = zip(*_hhlib_function_signatures)
        stripped_names = [name.split("_")[1] for name in function_names]
        self._library_mock.mock_add_spec(stripped_names, spec_set=True)

        # Start QMI context.
        self._ctx = QMI_Context("hydraharp_realtime_test")
        self._ctx.start()

        # Create instrument instance. Limit MAX calls to avoid problems.
        PicoQuant_HydraHarp400.MAX_EVENTS_PER_CALL = self.PATCHED_MAX_EVENTS_PER_CALL
        if self._testMethodName == "test_pending_events_limit":
            self._hydraharp = self._ctx.make_instrument(
                "hydraharp",
                PicoQuant_HydraHarp400,
                "1111111",
                max_pending_events=self.PATCHED_MAX_PENDING_EVENTS,
            )
        else:
            self._hydraharp = self._ctx.make_instrument(
                "hydraharp",
                PicoQuant_HydraHarp400,
                "1111111"
            )

        self._library_mock.GetLibraryVersion.return_value = 0
        self._library_mock.OpenDevice.return_value = 0

        string_buffer = ctypes.create_string_buffer(b"1111111")
        with patch("sys.platform", "linux1"), patch(
                "ctypes.create_string_buffer", return_value=string_buffer
        ):
            self._hydraharp.open()

        self._hydraharp.initialize("T3", "EXTERNAL")

    def tearDown(self):
        # Close instrument instance.
        self._hydraharp.close()
        # Stop QMI context.
        self._ctx.stop()
        self._ctx = None

    def test_realtime_histogram(self):
        """Test that real-time histogram updates even if event filter is used. Low resolution test."""
        sync_rate = ctypes.c_int(int(self.SYNC_FREQUENCY_5MHz))
        resolution = ctypes.c_double(25.0)
        sync_period = 1e12 / sync_rate.value
        num_events = 50000
        event_rate = 2.0e9
        # Generate a random set of events.
        events_in = gen_random_events(
            num_events=num_events,
            event_rate=event_rate,
            sync_rate=sync_rate.value,
            resolution=resolution.value,
            channels=[0, 1],
            seed=2345,
        )

        # Patch the readFifo() function.
        fifo_words_in = events_to_fifo(events_in, sync_period, resolution.value)
        done_event = threading.Event()
        self._library_mock.ReadFiFo.side_effect = make_patched_read_fifo(fifo_words_in, done_event)

        # Subscribe to real-time histograms.
        recv = QMI_SignalReceiver()
        self._hydraharp.sig_histogram.subscribe(recv)

        # Configure event filter to reject events on both channels.
        # This should not affect the histograms.
        self._hydraharp.set_event_filter(
            channel_filter={0: EventFilterMode.NO_EVENTS, 1: EventFilterMode.NO_EVENTS}
        )

        # Configure real-time histograms.
        bin_resolution = int(resolution.value)
        histogram_bins = 2**15 // bin_resolution  # Max resolution in T3 mode
        histogram_num_sync = int(
            event_rate // num_events // resolution.value
        )  # Important to be correct!
        self._hydraharp.set_realtime_histogram(
            channels=[0, 1],
            bin_resolution=bin_resolution,
            num_bins=histogram_bins,
            num_sync=histogram_num_sync,
        )

        # Start the measurement.
        with patch("ctypes.c_double", return_value=resolution), patch(
            "ctypes.c_int", return_value=sync_rate
        ):
            self._hydraharp.start_measurement(1000)

        # Wait to make sure all events are processed.
        done_event.wait()

        # Stop the measurement.
        self._hydraharp.stop_measurement()

        # Fetch events.
        events = self._hydraharp.get_events()

        # Expect only SYNC events.
        events_expected = events_in[(events_in["type"] == 64)]
        self.assertTrue(np.all(events == events_expected))

        # Check histograms.
        sync_count = 0
        last_sync = -1
        start_timestamp = 0
        for evt_type, evt_timestamp in events_in:
            if evt_type == 64:
                if sync_count % histogram_num_sync == 0:
                    if sync_count > 0:
                        # Check against published histogram.
                        sig = recv.get_next_signal()
                        (hist,) = sig.args
                        self.assertEqual(hist.start_timestamp, start_timestamp)
                        self.assertEqual(hist.bin_resolution, bin_resolution)
                        self.assertEqual(hist.num_sync, histogram_num_sync)
                        self.assertEqual(hist.channels, [0, 1])
                        self.assertTrue(np.all(hist.histogram_data == expect_histogram))

                    expect_histogram = np.zeros((2, histogram_bins), dtype=np.uint32)
                    start_timestamp = evt_timestamp

                sync_count += 1
                last_sync = int(evt_timestamp)

            elif sync_count > 0:
                # Count event in expected histogram.
                bin_index = min((int(evt_timestamp) - last_sync) // bin_resolution, histogram_bins - 1)
                expect_histogram[evt_type, bin_index] += 1

        # Check that there are no further published histograms.
        self.assertFalse(recv.has_signal_ready())

        self._hydraharp.sig_histogram.unsubscribe(recv)

    def test_realtime_countrate(self):
        """Test the countrate is correct despite of event block"""
        sync_rate = ctypes.c_int(int(self.SYNC_FREQUENCY_5MHz))
        resolution = ctypes.c_double(2.0)
        sync_period = 1e12 / sync_rate.value
        num_events = 50000
        event_rate = 2.0e9
        # Generate a random set of events.
        events_in = gen_random_events(
            num_events=num_events,
            event_rate=event_rate,
            sync_rate=sync_rate.value,
            resolution=resolution.value,
            channels=[0, 1],
            seed=2345,
        )

        # Patch the readFifo() function.
        fifo_words_in = events_to_fifo(events_in, sync_period, resolution.value)
        done_event = threading.Event()
        self._library_mock.ReadFiFo.side_effect = make_patched_read_fifo(fifo_words_in, done_event)

        # Subscribe to real-time count rate.
        recv = QMI_SignalReceiver()
        self._hydraharp.sig_countrate.subscribe(recv)

        # Block all events.
        # This should not affect the count rate reports.
        self._hydraharp.set_block_events(True)

        # Configure real-time count rates.
        delta_min = 25000
        delta_max = 75000
        countrate_num_sync = int(event_rate // num_events // resolution.value)  # Important to be correct!

        self._hydraharp.set_realtime_countrate(sync_aperture=(delta_min, delta_max), num_sync=countrate_num_sync)

        # Start the measurement.
        with patch("ctypes.c_double", return_value=resolution), patch(
            "ctypes.c_int", return_value=sync_rate
        ):
            self._hydraharp.start_measurement(1000)

        # Wait to make sure all events are processed.
        done_event.wait()

        # Stop the measurement.
        self._hydraharp.stop_measurement()

        # Fetch events.
        events = self._hydraharp.get_events()

        # Expect no events (all events blocked).
        self.assertEqual(len(events), 0)

        # Check count rates.
        sync_count = 0
        last_sync = -1
        start_timestamp = 0
        for evt_type, evt_timestamp in events_in:
            if evt_type == 64:
                if sync_count % countrate_num_sync == 0:
                    if sync_count > 0:
                        # Check against published count rate.
                        sig = recv.get_next_signal()
                        (counts,) = sig.args
                        self.assertEqual(counts.start_timestamp, start_timestamp)
                        self.assertEqual(counts.num_sync, countrate_num_sync)
                        self.assertTrue(np.all(counts.counts == expect_counts))

                    expect_counts = np.zeros(8, dtype=np.uint64)
                    start_timestamp = evt_timestamp

                sync_count += 1
                last_sync = int(evt_timestamp)

            elif sync_count > 0:
                # Check aperture filter.
                if delta_min <= int(evt_timestamp) - last_sync <= delta_max:
                    # Count this event.
                    expect_counts[evt_type] += 1

        # Check that there are no further published count rates.
        self.assertFalse(recv.has_signal_ready())

        self._hydraharp.sig_histogram.unsubscribe(recv)


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import patch

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.core.transport import create_transport
from qmi.instruments.anapico import Anapico_Apsin


class AnapicoApsinTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.transport_patch = patch("qmi.instruments.anapico.apsin.create_transport", autospec=create_transport)
        self.transport_patch.start()
        self.scpi_patch = patch("qmi.instruments.anapico.apsin.ScpiProtocol", autospec=ScpiProtocol)
        self.scpi_patch.start()
        self.apsin = Anapico_Apsin(QMI_Context("test_apsin"), name="apsin", transport="mock")

    def tearDown(self) -> None:
        self.transport_patch.stop()
        self.scpi_patch.stop()

    def test_open(self):
        """Test that the instrument can be opened."""
        # Act
        self.apsin.open()
        # Assert
        self.assertFalse(self.apsin._power_unit_configured)
        self.apsin._transport.open.assert_called_once_with()
        self.apsin._scpi_protocol.write.assert_called_once_with("*CLS")
        self.apsin._scpi_protocol.ask.assert_called_once_with("SYST:ERR:ALL?")

    def test_open_with_exception(self):
        """Force an exception at SCPI call and see it gets caught."""
        # Arrange
        self.apsin._scpi_protocol.write = Exception("Fail")
        # Act
        with self.assertRaises(Exception):
            self.apsin.open()

        self.apsin._transport.open.assert_called_once_with()
        self.apsin._transport.close.assert_called_once_with()

    def test_close(self):
        """Test the close call."""
        # Arrange
        self.apsin.open()
        # Act
        self.apsin.close()
        # Assert
        self.apsin._transport.close.assert_called_once_with()

    def test_reset(self):
        """Test the reset method."""
        # Arrange
        expected_write_calls = ["*CLS", "*RST"]
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.reset()
        # Assert
        self.apsin._scpi_protocol.ask.assert_called_with(expected_ask_call)
        self.assertEqual(2, self.apsin._scpi_protocol.ask.call_count)
        self.assertEqual(2, self.apsin._transport.discard_read.call_count)
        for call in expected_write_calls:
            self.apsin._scpi_protocol.write.assert_any_call(call)

    def test_check_error_excepts(self):
        """Test the case where a command returns an error and except in _check_error."""
        # Arrange
        error = "1: Some kind of error"
        expected_error_str = "Instrument returned error: {}".format(error)
        self.apsin._scpi_protocol.ask.side_effect = ["0, No error", error]
        # Act
        with self.assertRaises(QMI_InstrumentException) as exc:
            self.apsin.reset()

        self.assertEqual(expected_error_str, str(exc.exception))

    def test_get_idn(self):
        """Test the get_idn returns an instrument identification."""
        # Arrange
        vendor = "Vendor"
        model = "Model"
        serial = "Serial"
        version = "Version"
        self.apsin._scpi_protocol.ask.return_value = f"{vendor},{model},{serial},{version}\r\n"
        # Act
        idn = self.apsin.get_idn()
        # assert
        self.apsin._scpi_protocol.ask.assert_called_once_with("*IDN?")
        self.assertEqual(vendor, idn.vendor)
        self.assertEqual(model, idn.model)
        self.assertEqual(serial, idn.serial)
        self.assertEqual(version, idn.version)

    def test_get_idn_excepts(self):
        """Test that the get_idn excepts if not 4 words are returned."""
        # Arrange
        self.apsin._scpi_protocol.ask.side_effect = ["A,B,C\r\n", "1, 2, 3, 4, 5\r\n"]
        # Act
        for _ in range(2):
            with self.assertRaises(QMI_InstrumentException) as exc:
                self.apsin.get_idn()

            self.assertIn("Unexpected response to *IDN?", str(exc.exception))

    def test_get_output_enabled(self):
        """Test the get_output_enabled function."""
        # Arrange
        expected_call = ":OUTP?"
        self.apsin._scpi_protocol.ask.return_value = "1"
        # Act
        enabled = self.apsin.get_output_enabled()
        # Assert
        self.assertTrue(enabled)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_call)

    def test_ask_bool_excepts(self):
        """Test the _ask_bool function excepts at some invalid response with some command."""
        # Arrange
        expected_call = ":OUTP?"
        self.apsin._scpi_protocol.ask.return_value = "2"
        # Act
        with self.assertRaises(QMI_InstrumentException):
            self.apsin.get_output_enabled()

        # Assert
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_call)

    def test_set_output_enabled_enable(self):
        """Test set_output_enabled function with enabling."""
        # Arrange
        enable = True
        expected_write_call = f":OUTP {int(enable)}"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_output_enabled(enable)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_set_output_enabled_disable(self):
        """Test set_output_enabled function with disabling."""
        # Arrange
        enable = False
        expected_write_call = f":OUTP {int(enable)}"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_output_enabled(enable)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_get_frequency(self):
        """Test get_frequency function."""
        # Arrange
        expected_frequency = 123.45
        expected_ask_call = ":FREQ?"
        self.apsin._scpi_protocol.ask.return_value = f"{expected_frequency}\r\n"
        # Act
        freq = self.apsin.get_frequency()
        # Assert
        self.assertEqual(freq, expected_frequency)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_ask_float_excepts(self):
        """Test that the _ask_float excepts with invalid return value at some command."""
        # Arrange
        expected_ask_call = ":FREQ?"
        self.apsin._scpi_protocol.ask.return_value = f"off\r\n"
        # Act
        with self.assertRaises(QMI_InstrumentException):
            self.apsin.get_frequency()

        # Assert
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_set_frequency(self):
        """Test set_frequency call."""
        # Arrange
        frequency = 432.15
        expected_write_call = f":FREQ {frequency}"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_frequency(frequency)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_get_phase(self):
        # Arrange
        expected_phase = 3.141
        expected_ask_call = ":PHAS?"
        self.apsin._scpi_protocol.ask.return_value = f"{expected_phase}\r\n"
        # Act
        freq = self.apsin.get_phase()
        # Assert
        self.assertEqual(freq, expected_phase)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_set_phase(self):
        """Test set_phase call."""
        # Arrange
        phase = 0.0
        expected_write_call = f":PHAS {phase}"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_phase(phase)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_get_power(self):
        """Test get_power method."""
        # Arrange
        self.apsin._power_unit_configured = True  # First emulate already configured power unit
        expected_power = 99.999
        expected_write_call = ":UNIT:POW DBM"
        expected_ask_calls = [":POW?", ":SYST:ERR:ALL?"]
        self.apsin._scpi_protocol.ask.side_effect = [f"{expected_power}", "0, No error", f"{expected_power}"]
        # Act
        power_1 = self.apsin.get_power()
        # Assert
        self.assertEqual(expected_power, power_1)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_calls[0])
        self.apsin._scpi_protocol.write.assert_not_called()
        # Re-arrange
        self.apsin._power_unit_configured = False  # Then force to call for starting power unit first.
        # Second Act
        power_2 = self.apsin.get_power()
        # Assert
        self.assertTrue(self.apsin._power_unit_configured)
        self.assertEqual(expected_power, power_2)
        self.apsin._scpi_protocol.ask.assert_any_call(expected_ask_calls[1])
        self.assertEqual(3, self.apsin._scpi_protocol.ask.call_count)
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)

    def test_set_power(self):
        # Arrange
        self.apsin._power_unit_configured = True  # Emulate already configured power unit
        power = 99.999
        expected_write_call = f":POW {power}"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_power(power)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_get_pulsemod_enabled(self):
        """Test the get_pulsemod_enabled function."""
        # Arrange
        expected_call = ":PULM:STAT?"
        self.apsin._scpi_protocol.ask.return_value = "0"
        # Act
        enabled = self.apsin.get_pulsemod_enabled()
        # Assert
        self.assertFalse(enabled)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_call)

    def test_set_pulsemod_enabled_enable(self):
        """Test set_pulsemod_enabled function with enabling."""
        # Arrange
        enable = True
        expected_write_call = f":PULM:STAT {int(enable)}"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_pulsemod_enabled(enable)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_set_pulsemod_enabled_disable(self):
        """Test set_pulsemod_enabled function with disabling."""
        # Arrange
        enable = False
        expected_write_call = f":PULM:STAT {int(enable)}"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_pulsemod_enabled(enable)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_get_pulsemod_ext_source_is_ext(self):
        """Test the get_pulsemod_ext_source function returning "EXT"."""
        # Arrange
        expected_call = ":PULM:SOUR?"
        self.apsin._scpi_protocol.ask.return_value = "EXT"
        # Act
        ext_source = self.apsin.get_pulsemod_ext_source()
        # Assert
        self.assertTrue(ext_source)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_call)

    def test_get_pulsemod_ext_source_is_int(self):
        """Test the get_pulsemod_ext_source function returning "INT"."""
        # Arrange
        expected_call = ":PULM:SOUR?"
        self.apsin._scpi_protocol.ask.return_value = "INT"
        # Act
        ext_source = self.apsin.get_pulsemod_ext_source()
        # Assert
        self.assertFalse(ext_source)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_call)

    def test_get_pulsemod_ext_source_excepts(self):
        """Test the get_pulsemod_ext_source function excepts on wrong response."""
        # Arrange
        expected_call = ":PULM:SOUR?"
        self.apsin._scpi_protocol.ask.return_value = "OUT"
        # Act
        with self.assertRaises(QMI_InstrumentException):
            self.apsin.get_pulsemod_ext_source()

        # Assert
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_call)

    def test_set_pulsemod_ext_source_to_ext(self):
        """Test set_pulsemod_ext_source function with setting to "EXT"."""
        # Arrange
        ext = True
        expected_write_call = ":PULM:SOUR EXT"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_pulsemod_ext_source(ext)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_set_pulsemod_ext_source_to_int(self):
        """Test set_pulsemod_ext_source function with disabling."""
        # Arrange
        ext = False
        expected_write_call = ":PULM:SOUR INT"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_pulsemod_ext_source(ext)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_get_pulsemod_polarity_is_inv(self):
        """Test the get_pulsemod_polarity function returning "INV"."""
        # Arrange
        expected_call = ":PULM:POL?"
        self.apsin._scpi_protocol.ask.return_value = "INV"
        # Act
        polarity = self.apsin.get_pulsemod_polarity()
        # Assert
        self.assertTrue(polarity)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_call)

    def test_get_pulsemod_polarity_is_int(self):
        """Test the get_pulsemod_polarity function returning "NORM"."""
        # Arrange
        expected_call = ":PULM:POL?"
        self.apsin._scpi_protocol.ask.return_value = "NORM"
        # Act
        polarity = self.apsin.get_pulsemod_polarity()
        # Assert
        self.assertFalse(polarity)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_call)

    def test_get_pulsemod_polarity_excepts(self):
        """Test the get_pulsemod_polarity function excepts on wrong response."""
        # Arrange
        expected_call = ":PULM:POL?"
        self.apsin._scpi_protocol.ask.return_value = "0"
        # Act
        with self.assertRaises(QMI_InstrumentException):
            self.apsin.get_pulsemod_polarity()

        # Assert
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_call)

    def test_set_pulsemod_polarity_to_inv(self):
        """Test set_pulsemod_polarity function with setting to "INV"."""
        # Arrange
        inv = True
        expected_write_call = ":PULM:POL INV"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_pulsemod_polarity(inv)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_set_pulsemod_polarity_to_norm(self):
        """Test set_pulsemod_polarity function with disabling."""
        # Arrange
        inv = False
        expected_write_call = ":PULM:POL NORM"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_pulsemod_polarity(inv)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_get_am_enabled(self):
        """Test the get_am_enabled function."""
        # Arrange
        expected_call = ":AM:STAT?"
        self.apsin._scpi_protocol.ask.return_value = "ON"
        # Act
        enabled = self.apsin.get_am_enabled()
        # Assert
        self.assertTrue(enabled)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_call)

    def test_set_am_enabled_enable(self):
        """Test set_am_enabled function with enabling."""
        # Arrange
        enable = True
        expected_write_call = f":AM:STAT {int(enable)}"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_am_enabled(enable)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_set_am_enabled_disable(self):
        """Test set_am_enabled function with disabling."""
        # Arrange
        enable = False
        expected_write_call = f":AM:STAT {int(enable)}"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_am_enabled(enable)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_get_am_ext_source_is_ext(self):
        """Test the get_am_ext_source function returning "EXT"."""
        # Arrange
        expected_call = ":AM:SOUR?"
        self.apsin._scpi_protocol.ask.return_value = "EXT"
        # Act
        ext_source = self.apsin.get_am_ext_source()
        # Assert
        self.assertTrue(ext_source)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_call)

    def test_get_am_ext_source_is_int(self):
        """Test the get_am_ext_source function returning "INT"."""
        # Arrange
        expected_call = ":AM:SOUR?"
        self.apsin._scpi_protocol.ask.return_value = "INT"
        # Act
        ext_source = self.apsin.get_am_ext_source()
        # Assert
        self.assertFalse(ext_source)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_call)

    def test_set_am_ext_source_to_ext(self):
        """Test set_am_ext_source function with setting to "EXT"."""
        # Arrange
        ext = True
        expected_write_call = ":AM:SOUR EXT"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_am_ext_source(ext)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_set_am_ext_source_to_int(self):
        """Test set_am_ext_source function with setting to "INT"."""
        # Arrange
        ext = False
        expected_write_call = ":AM:SOUR INT"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_am_ext_source(ext)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_get_am_sensitivity(self):
        # Arrange
        expected_am_sensitivity = 3.141
        expected_ask_call = ":AM:SENS?"
        self.apsin._scpi_protocol.ask.return_value = f"{expected_am_sensitivity}\r\n"
        # Act
        freq = self.apsin.get_am_sensitivity()
        # Assert
        self.assertEqual(freq, expected_am_sensitivity)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_set_am_sensitivity(self):
        """Test set_am_sensitivity call."""
        # Arrange
        am_sensitivity = 0.0
        expected_write_call = f":AM:SENS {am_sensitivity}"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_am_sensitivity(am_sensitivity)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_get_fm_enabled(self):
        """Test the get_fm_enabled function."""
        # Arrange
        expected_call = ":FM:STAT?"
        self.apsin._scpi_protocol.ask.return_value = "OFF"
        # Act
        enabled = self.apsin.get_fm_enabled()
        # Assert
        self.assertFalse(enabled)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_call)

    def test_set_fm_enabled_enable(self):
        """Test set_fm_enabled function with enabling."""
        # Arrange
        enable = True
        expected_write_call = f":FM:STAT {int(enable)}"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_fm_enabled(enable)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_set_fm_enabled_disable(self):
        """Test set_fm_enabled function with disabling."""
        # Arrange
        enable = False
        expected_write_call = f":FM:STAT {int(enable)}"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_fm_enabled(enable)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_get_fm_ext_source_is_ext(self):
        """Test the get_fm_ext_source function returning "EXT"."""
        # Arrange
        expected_call = ":FM:SOUR?"
        self.apsin._scpi_protocol.ask.return_value = "EXT"
        # Act
        ext_source = self.apsin.get_fm_ext_source()
        # Assert
        self.assertTrue(ext_source)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_call)

    def test_get_fm_ext_source_is_int(self):
        """Test the get_fm_ext_source function returning "INT"."""
        # Arrange
        expected_call = ":FM:SOUR?"
        self.apsin._scpi_protocol.ask.return_value = "INT"
        # Act
        ext_source = self.apsin.get_fm_ext_source()
        # Assert
        self.assertFalse(ext_source)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_call)

    def test_set_fm_ext_source_to_ext(self):
        """Test set_fm_ext_source function with setting to "EXT"."""
        # Arrange
        ext = True
        expected_write_call = ":FM:SOUR EXT"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_fm_ext_source(ext)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_set_fm_ext_source_to_int(self):
        """Test set_fm_ext_source function with setting to "INT"."""
        # Arrange
        ext = False
        expected_write_call = ":FM:SOUR INT"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_fm_ext_source(ext)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_get_fm_sensitivity(self):
        # Arrange
        expected_fm_sensitivity = 3.141
        expected_ask_call = ":FM:SENS?"
        self.apsin._scpi_protocol.ask.return_value = f"{expected_fm_sensitivity}\r\n"
        # Act
        freq = self.apsin.get_fm_sensitivity()
        # Assert
        self.assertEqual(freq, expected_fm_sensitivity)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_set_fm_sensitivity(self):
        """Test set_fm_sensitivity call."""
        # Arrange
        fm_sensitivity = 0.0
        expected_write_call = f":FM:SENS {fm_sensitivity}"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_fm_sensitivity(fm_sensitivity)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_get_fm_coupling(self):
        """Test the get_fm_coupling function."""
        # Arrange
        expected_coupling = "AC"
        expected_call = ":FM:COUP?"
        self.apsin._scpi_protocol.ask.return_value = expected_coupling
        # Act
        coupling = self.apsin.get_fm_coupling()
        # Assert
        self.assertEqual(expected_coupling, coupling)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_call)

    def test_set_fm_coupling(self):
        """Test set_fm_coupling function."""
        # Arrange
        coupling = "dc"
        expected_write_call = f":FM:COUP {coupling.upper()}"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_fm_coupling(coupling)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_set_fm_coupling_excepts(self):
        """Test set_fm_coupling function excepts with invalid value."""
        # Arrange
        coupling = "Thunderstruck"
        # Act
        with self.assertRaises(ValueError):
            self.apsin.set_fm_coupling(coupling)

        self.apsin._scpi_protocol.write.assert_not_called()
        self.apsin._scpi_protocol.ask.assert_not_called()

    def test_get_reference_source(self):
        """Test the get_reference_source function."""
        # Arrange
        expected_source = "INT"
        expected_call = ":ROSC:SOUR?"
        self.apsin._scpi_protocol.ask.return_value = expected_source
        # Act
        source = self.apsin.get_reference_source()
        # Assert
        self.assertEqual(expected_source, source)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_call)

    def test_set_reference_source(self):
        """Test set_reference_source function."""
        # Arrange
        source = "int"
        expected_write_call = f":ROSC:SOUR {source.upper()}"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_reference_source(source)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_set_reference_source_excepts(self):
        """Test set_reference_source function excepts with invalid value."""
        # Arrange
        source = "ULT"
        # Act
        with self.assertRaises(ValueError):
            self.apsin.set_reference_source(source)

        self.apsin._scpi_protocol.write.assert_not_called()
        self.apsin._scpi_protocol.ask.assert_not_called()

    def test_get_external_reference_frequency(self):
        """Test get_external_reference_frequency function."""
        # Arrange
        expected_external_reference_frequency = 123.45E-5
        expected_ask_call = ":ROSC:EXT:FREQ?"
        self.apsin._scpi_protocol.ask.return_value = f"{expected_external_reference_frequency}\r\n"
        # Act
        freq = self.apsin.get_external_reference_frequency()
        # Assert
        self.assertEqual(freq, expected_external_reference_frequency)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_set_external_reference_frequency(self):
        """Test set_external_reference_frequency call."""
        # Arrange
        external_reference_frequency = 432.15E-6
        expected_write_call = f":ROSC:EXT:FREQ {external_reference_frequency}"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_external_reference_frequency(external_reference_frequency)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_get_reference_is_locked(self):
        """Test the get_reference_is_locked function."""
        # Arrange
        expected_call = ":ROSC:LOCK?"
        self.apsin._scpi_protocol.ask.return_value = "1"
        # Act
        locked = self.apsin.get_reference_is_locked()
        # Assert
        self.assertTrue(locked)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_call)

    def test_get_reference_output_enabled(self):
        """Test the get_reference_output_enabled function."""
        # Arrange
        expected_call = ":ROSC:OUTP?"
        self.apsin._scpi_protocol.ask.return_value = "OFF"
        # Act
        enabled = self.apsin.get_reference_output_enabled()
        # Assert
        self.assertFalse(enabled)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_call)

    def test_set_reference_output_enabled_enable(self):
        """Test set_reference_output_enabled function with enabling."""
        # Arrange
        enable = True
        expected_write_call = f":ROSC:OUTP {int(enable)}"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_reference_output_enabled(enable)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)

    def test_set_reference_output_enabled_disable(self):
        """Test set_reference_output_enabled function with disabling."""
        # Arrange
        enable = False
        expected_write_call = f":ROSC:OUTP {int(enable)}"
        expected_ask_call = ":SYST:ERR:ALL?"
        self.apsin._scpi_protocol.ask.return_value = "0, No error"
        # Act
        self.apsin.set_reference_output_enabled(enable)
        # Assert
        self.apsin._scpi_protocol.write.assert_called_once_with(expected_write_call)
        self.apsin._scpi_protocol.ask.assert_called_once_with(expected_ask_call)


if __name__ == '__main__':
    unittest.main()



"""A Python wrapper module to wrap around manufacturer provided library header file 'phdefin.h'. The header file used
is from software version 3.0.0.3.
Please find the header file in the dedicated software package for the PicoHarp instrument at
https://www.picoquant.com/dl_software/PicoHarp300/PicoHarp300_SW_and_DLL_v3_0_0_3.zip"""
# This is generated code, do not edit by hand!
import ctypes


_phlib_function_signatures = [
        ('PH_GetLibraryVersion', ctypes.c_int, [('version', ctypes.POINTER(ctypes.c_char))]),
        ('PH_GetErrorString', ctypes.c_int, [('errstring', ctypes.POINTER(ctypes.c_char)), ('errcode', ctypes.c_int)]),
        ('PH_OpenDevice', ctypes.c_int, [('devidx', ctypes.c_int), ('serial', ctypes.POINTER(ctypes.c_char))]),
        ('PH_CloseDevice', ctypes.c_int, [('devidx', ctypes.c_int)]),
        ('PH_Initialize', ctypes.c_int, [('devidx', ctypes.c_int), ('mode', ctypes.c_int)]),
        ('PH_GetHardwareInfo', ctypes.c_int, [('devidx', ctypes.c_int), ('model', ctypes.POINTER(ctypes.c_char)), ('partno', ctypes.POINTER(ctypes.c_char)), ('version', ctypes.POINTER(ctypes.c_char))]),
        ('PH_GetSerialNumber', ctypes.c_int, [('devidx', ctypes.c_int), ('serial', ctypes.POINTER(ctypes.c_char))]),
        ('PH_GetFeatures', ctypes.c_int, [('devidx', ctypes.c_int), ('features', ctypes.POINTER(ctypes.c_int))]),
        ('PH_GetBaseResolution', ctypes.c_int, [('devidx', ctypes.c_int), ('resolution', ctypes.POINTER(ctypes.c_double)), ('binsteps', ctypes.POINTER(ctypes.c_int))]),
        ('PH_GetHardwareDebugInfo', ctypes.c_int, [('devidx', ctypes.c_int), ('debuginfo', ctypes.POINTER(ctypes.c_char))]),
        ('PH_Calibrate', ctypes.c_int, [('devidx', ctypes.c_int)]),
        ('PH_SetInputCFD', ctypes.c_int, [('devidx', ctypes.c_int), ('channel', ctypes.c_int), ('level', ctypes.c_int), ('zc', ctypes.c_int)]),
        ('PH_SetSyncDiv', ctypes.c_int, [('devidx', ctypes.c_int), ('div', ctypes.c_int)]),
        ('PH_SetSyncOffset', ctypes.c_int, [('devidx', ctypes.c_int), ('syncoffset', ctypes.c_int)]),
        ('PH_SetStopOverflow', ctypes.c_int, [('devidx', ctypes.c_int), ('stop_ovfl', ctypes.c_int), ('stopcount', ctypes.c_int)]),
        ('PH_SetBinning', ctypes.c_int, [('devidx', ctypes.c_int), ('binning', ctypes.c_int)]),
        ('PH_SetOffset', ctypes.c_int, [('devidx', ctypes.c_int), ('offset', ctypes.c_int)]),
        ('PH_SetMultistopEnable', ctypes.c_int, [('devidx', ctypes.c_int), ('enable', ctypes.c_int)]),
        ('PH_ClearHistMem', ctypes.c_int, [('devidx', ctypes.c_int), ('block', ctypes.c_int)]),
        ('PH_StartMeas', ctypes.c_int, [('devidx', ctypes.c_int), ('tacq', ctypes.c_int)]),
        ('PH_StopMeas', ctypes.c_int, [('devidx', ctypes.c_int)]),
        ('PH_CTCStatus', ctypes.c_int, [('devidx', ctypes.c_int), ('ctcstatus', ctypes.POINTER(ctypes.c_int))]),
        ('PH_GetHistogram', ctypes.c_int, [('devidx', ctypes.c_int), ('chcount', ctypes.POINTER(ctypes.c_uint)), ('block', ctypes.c_int)]),
        ('PH_GetResolution', ctypes.c_int, [('devidx', ctypes.c_int), ('resolution', ctypes.POINTER(ctypes.c_double))]),
        ('PH_GetCountRate', ctypes.c_int, [('devidx', ctypes.c_int), ('channel', ctypes.c_int), ('rate', ctypes.POINTER(ctypes.c_int))]),
        ('PH_GetFlags', ctypes.c_int, [('devidx', ctypes.c_int), ('flags', ctypes.POINTER(ctypes.c_int))]),
        ('PH_GetElapsedMeasTime', ctypes.c_int, [('devidx', ctypes.c_int), ('elapsed', ctypes.POINTER(ctypes.c_double))]),
        ('PH_GetWarnings', ctypes.c_int, [('devidx', ctypes.c_int), ('warnings', ctypes.POINTER(ctypes.c_int))]),
        ('PH_GetWarningsText', ctypes.c_int, [('devidx', ctypes.c_int), ('text', ctypes.POINTER(ctypes.c_char)), ('warnings', ctypes.c_int)]),
        ('PH_SetMarkerEnable', ctypes.c_int, [('devidx', ctypes.c_int), ('en0', ctypes.c_int), ('en1', ctypes.c_int), ('en2', ctypes.c_int), ('en3', ctypes.c_int)]),
        ('PH_SetMarkerEdges', ctypes.c_int, [('devidx', ctypes.c_int), ('me0', ctypes.c_int), ('me1', ctypes.c_int), ('me2', ctypes.c_int), ('me3', ctypes.c_int)]),
        ('PH_SetMarkerHoldoffTime', ctypes.c_int, [('devidx', ctypes.c_int), ('holdofftime', ctypes.c_int)]),
        ('PH_ReadFiFo', ctypes.c_int, [('devidx', ctypes.c_int), ('buffer', ctypes.POINTER(ctypes.c_uint)), ('count', ctypes.c_int), ('nactual', ctypes.POINTER(ctypes.c_int))]),
        ('PH_GetRouterVersion', ctypes.c_int, [('devidx', ctypes.c_int), ('model', ctypes.POINTER(ctypes.c_char)), ('version', ctypes.POINTER(ctypes.c_char))]),
        ('PH_GetRoutingChannels', ctypes.c_int, [('devidx', ctypes.c_int), ('rtchannels', ctypes.POINTER(ctypes.c_int))]),
        ('PH_EnableRouting', ctypes.c_int, [('devidx', ctypes.c_int), ('enable', ctypes.c_int)]),
        ('PH_SetRoutingChannelOffset', ctypes.c_int, [('devidx', ctypes.c_int), ('channel', ctypes.c_int), ('offset', ctypes.c_int)]),
        ('PH_SetPHR800Input', ctypes.c_int, [('devidx', ctypes.c_int), ('channel', ctypes.c_int), ('level', ctypes.c_int), ('edge', ctypes.c_int)]),
        ('PH_SetPHR800CFD', ctypes.c_int, [('devidx', ctypes.c_int), ('channel', ctypes.c_int), ('level', ctypes.c_int), ('zc', ctypes.c_int)])
    ]



import os
import tempfile
import unittest
import unittest.mock
from unittest.mock import patch
import pathlib

import numpy as np

from qmi.core.exceptions import QMI_InstrumentException, QMI_ConfigurationException
from qmi.instruments.adwin.adwin import Adwin_Base
from qmi.utils.adwin_manager import AdwinProgramLibrary, AdwinProcess, AdwinManager, ProgramInfo
from qmi.utils.adbasic_compiler import AdbasicCompilerException, AdbasicError

CONFIG_BASE = """# Comment
{
    # File name of the top-level ".bas" file without the ".bas" extension.
    "file": "program_key_error",

    # Process slot number on the ADwin (range 1 to 10).
    "slot": 9,

    # Event trigger source (either "timer" or "external").
    "trigger": "external",

    # Process priority (integer between -10 and +10 for low priority, or 1000 for high priority).
    "priority": 1000,

    # True to discover parameters by parsing the ADbasic source code.
    "parse_parameters": true
}
"""


def write_source_file(tempdir, file_name, content):
    with open(os.path.join(tempdir, file_name), "w") as outf:
        outf.write(content)


class AdwinProgramLibraryTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self._test_data_dir = os.path.join(os.path.dirname(__file__), 'data')
        # Load adwin program library with default parameters; processor = T12, HW = PII
        self._adwin_program_library = AdwinProgramLibrary(self._test_data_dir, self._test_data_dir)

    def test_list_programs(self):
        # arrange
        expected_program_list = ["program_a", "program_b", "program_c", "program_d", "program_e"]

        # act
        program_list = self._adwin_program_library.list_programs()

        # assert
        self.assertTrue(all(program in expected_program_list for program in program_list))

    def test_get_program_info(self):
        """Get program info from configuration for loaded programs"""
        # act
        prog_a_info = self._adwin_program_library.get_program_info("program_a")
        param_a_info = prog_a_info.param_info
        prog_b_info = self._adwin_program_library.get_program_info("program_b")
        param_b_info = prog_b_info.param_info
        prog_c_info = self._adwin_program_library.get_program_info("program_c")
        param_c_info = prog_c_info.param_info

        # assert
        self.assertIn("foo", param_a_info.param.keys())
        self.assertIn("bar", param_a_info.param.keys())
        self.assertIn("foobar", param_b_info.param.keys())
        self.assertIn("foo", param_c_info.param.keys())
        self.assertIn("bar", param_c_info.param.keys())

    def test_get_program_info_from_config_file(self):
        """Now test the same from the file using the ProgramInfo class method."""
        # act
        prog_a_conf = os.path.join(self._test_data_dir, "program_a.conf")
        prog_a_info = ProgramInfo.from_config_file(prog_a_conf, self._test_data_dir)
        param_a_info = prog_a_info.param_info
        prog_b_conf = os.path.join(self._test_data_dir, "program_b.conf")
        prog_b_info = ProgramInfo.from_config_file(prog_b_conf, self._test_data_dir)  # , processor_type="T12.1")
        param_b_info = prog_b_info.param_info
        prog_c_conf = os.path.join(self._test_data_dir, "program_c.conf")
        prog_c_info = ProgramInfo.from_config_file(prog_c_conf, self._test_data_dir)  # , processor_type="T11")
        param_c_info = prog_c_info.param_info

        # assert
        self.assertIn("foo", param_a_info.param.keys())
        self.assertIn("bar", param_a_info.param.keys())
        self.assertIn("foobar", param_b_info.param.keys())
        self.assertIn("foo", param_c_info.param.keys())
        self.assertIn("bar", param_c_info.param.keys())

    def test_get_program_info_false_parse_parameters(self):
        """parse_parameters is now set as 'false', so specific read of 'par', 'fpar', and 'par_array' will be done"""
        # act
        prog_d_info = self._adwin_program_library.get_program_info("program_d")
        param_d_info = prog_d_info.param_info

        # assert
        self.assertIn("uno", param_d_info.param.keys())
        self.assertIn("due.zero", param_d_info.param.keys())
        self.assertIn("tres_quattro", param_d_info.param.keys())

    def test_get_program_info_twice(self):
        """Second 'get' should return info immediately without reading config"""
        # act
        prog_d_info = self._adwin_program_library.get_program_info("program_d")
        prog_d_info_again = self._adwin_program_library.get_program_info("program_d")

        # assert
        self.assertEqual(prog_d_info, prog_d_info_again)

    def test_get_program_info_raises_exception(self):
        """Should raise exception due to double variable name 'uno'"""
        # assert
        with self.assertRaises(QMI_ConfigurationException):
            self._adwin_program_library.get_program_info("program_e")

    @patch("qmi.utils.adwin_manager.compile_program")
    def test_compile_fail(self, patser):
        """Test the compile function fails at compiling the fake program, as expected"""
        # arrange
        message = "no real message"
        error = AdbasicError(
            error_number=2,
            error_description="test error",
            error_line="well I wanted it to err",
            filename="this file",
            line_number=5
        )
        patser.side_effect = AdbasicCompilerException(message, [error])
        patser.start()
        # act
        with self.assertRaises(AdbasicCompilerException) as ad_err:
            self._adwin_program_library.compile("program_d")

        self.assertEqual(ad_err.exception.adbasic_errors, [error])
        self.assertEqual(ad_err.exception.message, message)
        patser.stop()

    def test_config_has_wrong_key(self):
        """Test that wrong key in the config file raises exception."""
        file_name = "bas.conf"
        expected_exc = f"Error reading Adwin program config {file_name}"
        config_with_wrong_key = CONFIG_BASE.replace("slot", "plot")
        with tempfile.TemporaryDirectory(prefix="bastest") as tempdir:
            with self.assertRaises(QMI_ConfigurationException) as exc:
                write_source_file(tempdir, file_name, config_with_wrong_key)
                AdwinProgramLibrary(tempdir, tempdir)

            self.assertEqual(str(exc.exception), expected_exc)

    def test_config_has_wrong_slot_number(self):
        """Test that wrong slot number in the config file raises exception."""
        file_name = "bas.conf"
        expected_exc = "Invalid ADwin process slot number 0 for program 'bas'"
        config_with_wrong_key = CONFIG_BASE.replace('"slot": 9', '"slot": 0')
        with tempfile.TemporaryDirectory(prefix="bastest") as tempdir:
            with self.assertRaises(QMI_ConfigurationException) as exc:
                write_source_file(tempdir, file_name, config_with_wrong_key)
                AdwinProgramLibrary(tempdir, tempdir)

            self.assertEqual(str(exc.exception), expected_exc)

    def test_config_has_wrong_trigger(self):
        """Test that wrong slot number in the config file raises exception."""
        file_name = "bas.conf"
        expected_exc = "Invalid ADwin process trigger 'internal' for program 'bas'"
        config_with_wrong_key = CONFIG_BASE.replace('"trigger": "external"', '"trigger": "internal"')
        with tempfile.TemporaryDirectory(prefix="bastest") as tempdir:
            with self.assertRaises(QMI_ConfigurationException) as exc:
                write_source_file(tempdir, file_name, config_with_wrong_key)
                AdwinProgramLibrary(tempdir, tempdir)

            self.assertEqual(str(exc.exception), expected_exc)

    def test_config_has_wrong_priority_number(self):
        """Test that wrong priority number in the config file raises exception."""
        file_name = "bas.conf"
        expected_exc = "Invalid ADwin process priority 100 for program 'bas'"
        config_with_wrong_key = CONFIG_BASE.replace('"priority": 1000', '"priority": 100')
        with tempfile.TemporaryDirectory(prefix="bastest") as tempdir:
            with self.assertRaises(QMI_ConfigurationException) as exc:
                write_source_file(tempdir, file_name, config_with_wrong_key)
                AdwinProgramLibrary(tempdir, tempdir)

            self.assertEqual(str(exc.exception), expected_exc)

    def test_config_has_wrong_parameters(self):
        """Test that wrong slot number in the config file raises exception."""
        file_name = "bas.conf"
        expected_exc = "Invalid configuration for ADwin program 'bas': " + \
                       "Can not parse source code and specify explicit parameters"
        config_with_wrong_key = CONFIG_BASE[:-3] + ',\n\n\t"par": {"uno": 1}\n}'
        with tempfile.TemporaryDirectory(prefix="bastest") as tempdir:
            with self.assertRaises(QMI_ConfigurationException) as exc:
                write_source_file(tempdir, file_name, config_with_wrong_key)
                AdwinProgramLibrary(tempdir, tempdir)

            self.assertEqual(str(exc.exception), expected_exc)


class AdwinProcessTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self._test_data_dir = os.path.join(pathlib.Path(__file__).parent, 'data')

        self._adwin_program_library = AdwinProgramLibrary(self._test_data_dir, self._test_data_dir)
        self._adwin_program_library_b = AdwinProgramLibrary(self._test_data_dir, self._test_data_dir, "T12.1")
        self._adwin_program_library_d = AdwinProgramLibrary(self._test_data_dir, self._test_data_dir, "T11", "GII")

        self._adwin_mock = unittest.mock.MagicMock(spec=Adwin_Base)
        self._adwin_mock.is_process_running.return_value = False

        proc_a_info = self._adwin_program_library.get_program_info("program_a")
        proc_b_info = self._adwin_program_library_b.get_program_info("program_b")
        proc_d_info = self._adwin_program_library_d.get_program_info("program_d")

        self._adwin_mock.get_processor_type.return_value = "T12"
        self._adwin_proc = AdwinProcess(self._adwin_mock, "program_a", proc_a_info)
        self._adwin_mock.get_processor_type.return_value = "T12.1"
        self._adwin_proc_b = AdwinProcess(self._adwin_mock, "program_b", proc_b_info)
        self._adwin_mock.get_processor_type.return_value = "T11"
        self._adwin_proc_d = AdwinProcess(self._adwin_mock, "program_d", proc_d_info)

    def test_program_library_instances(self):
        """Test that the hardware and processor types are set correctly in the different program libraries."""
        expected_a = ["T12", "PII"]  # Defaults
        expected_b = ["T12.1", "PII"]
        expected_d = ["T11", "GII"]

        check_list_a = [self._adwin_program_library._processor_type, self._adwin_program_library._hardware_type]
        check_list_b = [self._adwin_program_library_b._processor_type, self._adwin_program_library_b._hardware_type]
        check_list_d = [self._adwin_program_library_d._processor_type, self._adwin_program_library_d._hardware_type]

        self.assertListEqual(expected_a, check_list_a)
        self.assertListEqual(expected_b, check_list_b)
        self.assertListEqual(expected_d, check_list_d)

        self.assertEqual(expected_a[0], self._adwin_proc._processor_type)
        self.assertEqual(expected_b[0], self._adwin_proc_b._processor_type)
        self.assertEqual(expected_d[0], self._adwin_proc_d._processor_type)

    def test_load(self):
        """Test loading without starting process"""
        file = os.path.join(pathlib.Path(__file__).parent, "data", "program_a")
        slot = 9
        expected_load_bin_file = f"{file}" + ".TC{}".format(slot % 10)
        self._adwin_proc.load()
        self._adwin_mock.load_process.assert_called_once_with(expected_load_bin_file)

    def test_load_b(self):
        """Test loading without starting process, with T12.1 processor"""
        file = os.path.join(pathlib.Path(__file__).parent, "data", "program_b")
        slot = 10
        expected_load_bin_file = f"{file}" + ".TC{}".format(slot % 10)
        self._adwin_proc_b.load()
        self._adwin_mock.load_process.assert_called_once_with(expected_load_bin_file)

    def test_load_d(self):
        """Test loading without starting process, with T11 processor"""
        file = os.path.join(pathlib.Path(__file__).parent, "data", "program_d")
        slot = 10
        expected_load_bin_file = f"{file}" + ".TB{}".format(slot % 10)
        self._adwin_proc_d.load()
        self._adwin_mock.load_process.assert_called_once_with(expected_load_bin_file)

    def test_start_with_params(self):
        self._adwin_proc.start_with_params(foo=10, Bar=11)

        self._adwin_mock.is_process_running.assert_called_once_with(9)

        self._adwin_mock.set_par.assert_has_calls([unittest.mock.call(1, 10), unittest.mock.call(2, 11)])

    def test_dont_start_with_params_if_running(self):
        self._adwin_mock.is_process_running.return_value = True
        with self.assertRaises(QMI_InstrumentException):
            self._adwin_proc.start_with_params(foo=10, Bar=11)

    def test_start(self):
        self._adwin_proc.start()
        self._adwin_mock.is_process_running.assert_called_once_with(9)

    def test_dont_start_if_running(self):
        self._adwin_mock.is_process_running.return_value = True
        with self.assertRaises(QMI_InstrumentException):
            self._adwin_proc.start()

    def test_stop(self):
        """Test stopping a process"""
        self._adwin_proc.stop()
        self._adwin_mock.stop_process.assert_called_once()

    def test_is_running(self):
        """Test the 'is_running' call"""
        is_it = self._adwin_proc.is_running()
        self.assertFalse(is_it)
        self._adwin_mock.is_process_running.assert_called_once_with(9)

    def test_wait_for_process(self):
        """Test the 'wait_for_process' call"""
        timeout = 0.01
        self._adwin_proc.wait_for_process(timeout)
        self._adwin_mock.wait_for_process.assert_called_once_with(9, timeout)

    def test_get_par(self):
        """Test getting parameters"""
        self._adwin_mock.get_par.side_effect = [10, 100]
        self._adwin_mock.get_fpar.side_effect = [1.0]
        self._adwin_mock.get_data.side_effect = [np.array([1000])]

        self._adwin_proc.start()
        par_foo = self._adwin_proc.get_par('foo')
        par_bar = self._adwin_proc.get_par('bar')
        par_baz = self._adwin_proc.get_par('baz')
        par_elem_boo = self._adwin_proc.get_par('elem_boo')

        self.assertEqual(par_foo, 10)
        self.assertEqual(par_bar, 100)
        self.assertAlmostEqual(par_baz, 1.0)
        self.assertEqual(par_elem_boo, 1000)

    def test_get_par_case_insensitive(self):
        self._adwin_mock.get_par.side_effect = [10, 100]
        self._adwin_mock.get_fpar.side_effect = [1.0]
        self._adwin_mock.get_data.side_effect = [np.array([1000])]

        self._adwin_proc.start()
        par_foo = self._adwin_proc.get_par('Foo')
        par_bar = self._adwin_proc.get_par('BAR')
        par_baz = self._adwin_proc.get_par('baz')
        par_elem_boo = self._adwin_proc.get_par('elem_boo')

        self.assertEqual(par_foo, 10)
        self.assertEqual(par_bar, 100)
        self.assertAlmostEqual(par_baz, 1.0)
        self.assertEqual(par_elem_boo, 1000)

    def test_set_par(self):
        self._adwin_proc.start()
        self._adwin_proc.set_par('Foo', 100)
        self._adwin_proc.set_par('BAR', 10)
        self._adwin_proc.set_par('baz', 2.0)
        self._adwin_proc.set_par('eLeM_BoO', 8)

        self._adwin_mock.set_par.assert_has_calls([unittest.mock.call(1, 100), unittest.mock.call(2, 10)])
        self._adwin_mock.set_fpar.assert_has_calls([unittest.mock.call(5, 2.0)])
        self._adwin_mock.set_data.assert_has_calls([unittest.mock.call(10, 1., np.array([8]))])

    def test_set_par_case_insensitive(self):
        self._adwin_proc.start()
        self._adwin_proc.set_par('Foo', 100)
        self._adwin_proc.set_par('BAR', 10)
        self._adwin_proc.set_par('baz', 2.0)
        self._adwin_proc.set_par('eLeM_BoO', 8)

        self._adwin_mock.set_par.assert_has_calls([unittest.mock.call(1, 100), unittest.mock.call(2, 10)])
        self._adwin_mock.set_fpar.assert_has_calls([unittest.mock.call(5, 2.0)])
        self._adwin_mock.set_data.assert_has_calls([unittest.mock.call(10, 1., np.array([8]))])

    def test_set_par_fail_if_set_float_for_int_par(self):
        self._adwin_proc.start()
        with self.assertRaises(TypeError):
            self._adwin_proc.set_par('Foo', 1.0)

    def test_set_par_raise_exception_when_invalid_key(self):
        self._adwin_proc.start()
        with self.assertRaises(ValueError):
            self._adwin_proc.set_par('non_existing_par', 1.0)

    def test_get_par_multiple(self):
        self._adwin_mock.get_par.side_effect = [10, 100]
        self._adwin_mock.get_par.side_effect = [10, 100]
        self._adwin_mock.get_fpar.side_effect = [1.0]
        self._adwin_mock.get_data.side_effect = [np.array([1000])]

        self._adwin_proc.start()
        params = self._adwin_proc.get_par_multiple(['foo', 'bar', 'baz', 'elem_boo'])

        self.assertDictEqual(params, {'foo': 10, 'bar': 100, 'baz': 1.0, 'elem_boo': 1000})

    def test_get_par_multiple_case_insensitive(self):
        self._adwin_mock.get_par.side_effect = [10, 100]
        self._adwin_mock.get_par.side_effect = [10, 100]
        self._adwin_mock.get_fpar.side_effect = [1.0]
        self._adwin_mock.get_data.side_effect = [np.array([1000])]

        self._adwin_proc.start()
        params = self._adwin_proc.get_par_multiple(['foo', 'Bar', 'baz', 'elem_BOO'])

        self.assertDictEqual(params, {'foo': 10, 'Bar': 100, 'baz': 1.0, 'elem_BOO': 1000})

    def test_get_par_multiple_case_exception(self):
        """Test that invalid parameter raises ValueError"""
        self._adwin_mock.get_par.side_effect = [10, 100]
        self._adwin_mock.get_par.side_effect = [10, 100]
        self._adwin_mock.get_fpar.side_effect = [1.0]
        self._adwin_mock.get_data.side_effect = [np.array([1000])]

        self._adwin_proc.start()
        with self.assertRaises(ValueError):
            self._adwin_proc.get_par_multiple(['goo'])

    def test_set_par_multiple(self):
        self._adwin_proc.start()
        params = {'foo': 10, 'bar': 100, 'baz': 2.0, 'elem_boo': 8}
        self._adwin_proc.set_par_multiple(params)

        self._adwin_mock.set_par.assert_has_calls([unittest.mock.call(1, 10), unittest.mock.call(2, 100)])
        self._adwin_mock.set_fpar.assert_has_calls([unittest.mock.call(5, 2.0)])
        self._adwin_mock.set_data.assert_has_calls([unittest.mock.call(10, 1., np.array([8]))])

    def test_set_par_multiple_case_insensitive(self):
        self._adwin_proc.start()
        params = {'foo': 10, 'Bar': 100, 'baz': 2.0, 'eLeM_BoO': 8}
        self._adwin_proc.set_par_multiple(params)

        self._adwin_mock.set_par.assert_has_calls([unittest.mock.call(1, 10), unittest.mock.call(2, 100)])
        self._adwin_mock.set_fpar.assert_has_calls([unittest.mock.call(5, 2.0)])
        self._adwin_mock.set_data.assert_has_calls([unittest.mock.call(10, 1., np.array([8]))])

    def test_set_par_multiple_fails_when_float_for_int_par(self):
        self._adwin_proc.start()
        params = {'foo': 1.0}
        with self.assertRaises(TypeError):
            self._adwin_proc.set_par_multiple(params)

    def test_set_par_multiple_invalid_parameter(self):
        self._adwin_proc.start()
        params = {'goo': 1.0}
        with self.assertRaises(ValueError):
            self._adwin_proc.set_par_multiple(params)

    def test_get_data_length(self):
        self._adwin_mock.get_data_length.side_effect = [100]

        self._adwin_proc.start()
        length = self._adwin_proc.get_data_length('boo')

        self.assertEqual(length, 100)

    def test_get_data(self):
        expected_data = np.array([1, 2, 3, 4])
        self._adwin_mock.get_data.side_effect = [expected_data]

        self._adwin_proc.start()
        data = self._adwin_proc.get_data('boo', 1, 4)

        np.testing.assert_array_equal(data, expected_data)

    def test_get_full_data(self):
        expected_data = np.array([1, 2, 3, 4])
        self._adwin_mock.get_full_data.side_effect = [expected_data]

        self._adwin_proc.start()
        data = self._adwin_proc.get_full_data('boo')

        np.testing.assert_array_equal(data, expected_data)

    def test_set_data(self):
        self._adwin_proc.start()
        data = np.array([1, 2, 3, 4])
        self._adwin_proc.set_data('boo', 1, data)

        self._adwin_mock.set_data.assert_has_calls([unittest.mock.call(10, 1, data)])

    def test_get_fifo_filled(self):
        self._adwin_mock.get_fifo_filled.return_value = 10

        self._adwin_proc.start()
        fifo_room = self._adwin_proc.get_fifo_filled('boo')

        self.assertEqual(fifo_room, 10)

    def test_get_fifo_room(self):
        self._adwin_mock.get_fifo_room.return_value = 10

        self._adwin_proc.start()
        fifo_room = self._adwin_proc.get_fifo_room('boo')

        self.assertEqual(fifo_room, 10)

    def test_read_fifo(self):
        expected_fifo = np.array([1, 2, 3, 4])
        self._adwin_mock.read_fifo.return_value = expected_fifo

        self._adwin_proc.start()
        fifo = self._adwin_proc.read_fifo('boo', 4)

        np.testing.assert_array_equal(fifo, expected_fifo)

    def test_write_fifo(self):
        self._adwin_proc.start()
        data = np.array([1, 2, 3, 4])
        self._adwin_proc.write_fifo('boo', data)

        self._adwin_mock.write_fifo.assert_has_calls([unittest.mock.call(10, data)])


class TestAdwinManagerTestCase(unittest.TestCase):

    def setUp(self):
        self._test_data_dir = os.path.join(os.path.dirname(__file__), 'data')
        # adwin library gets the default processor type "T12" and harware type "PII".
        self._adwin_program_library = AdwinProgramLibrary(self._test_data_dir, self._test_data_dir)

        self._adwin_mock = unittest.mock.MagicMock(spec=Adwin_Base)
        self._adwin_mock.is_process_running.return_value = False

    def test_duplicate_auto_load_program_exception(self):
        """See that same process number cannot be used twice to load programs"""
        # Arrange
        expected = "Duplicate auto_load program in process slot number 10 (program_c)"
        auto_load_programs = ["program_a", "program_b", "program_c"]
        # Act & Assert
        with self.assertRaises(ValueError) as exc:
            AdwinManager(self._adwin_mock, self._adwin_program_library, auto_load_programs)

        self.assertEqual(exc.exception.args[0], expected)

    def test_auto_load_unknown_program_exception(self):
        """See that unknown program name raises exception"""
        # Arrange
        expected = "Unknown program 'program_x' in auto_load_programs"
        auto_load_programs = ["program_a", "program_b", "program_x"]
        # Act & Assert
        with self.assertRaises(ValueError) as exc:
            AdwinManager(self._adwin_mock, self._adwin_program_library, auto_load_programs)

        self.assertEqual(exc.exception.args[0], expected)

    def test_get_adwin(self):
        """Test that we get the same adwin object using the `get_adwin` method"""
        # Arrange
        auto_load_programs = ["program_a", "program_b"]
        self._adwin_manager = AdwinManager(self._adwin_mock, self._adwin_program_library, auto_load_programs)

        # Act
        adwin_copy = self._adwin_manager.get_adwin()

        # Assert
        self.assertIs(self._adwin_mock, adwin_copy)

    def test_get_program_library(self):
        """Test that we get the same program library object using the `get_program_library` method"""
        # Arrange
        auto_load_programs = ["program_a", "program_b"]
        self._adwin_manager = AdwinManager(self._adwin_mock, self._adwin_program_library, auto_load_programs)

        # Act
        program_library_copy = self._adwin_manager.get_program_library()

        # Assert
        self.assertIs(self._adwin_program_library, program_library_copy)

    def test_reboot(self):
        """Test that reboot reloads the program libraries"""
        # Arrange
        auto_load_programs = ["program_a", "program_b"]
        self._adwin_manager = AdwinManager(self._adwin_mock, self._adwin_program_library, auto_load_programs)
        program_list_expected = self._adwin_manager.list_programs()

        # Act
        self._adwin_manager.reboot()

        # Assert
        self.assertListEqual(program_list_expected, self._adwin_manager.list_programs())

    def test_get_process(self):
        """Test that we generate an AdwinProcess instance with `get_process` method"""
        # Arrange
        expected_instance_type = AdwinProcess
        auto_load_programs = ["program_a", "program_b"]
        self._adwin_manager = AdwinManager(self._adwin_mock, self._adwin_program_library, auto_load_programs)

        # Act
        adwin_process_copy = self._adwin_manager.get_process("program_a")

        # Assert
        self.assertIsInstance(adwin_process_copy, expected_instance_type)


if __name__ == '__main__':
    unittest.main()



"""
© 2024 HighFinesse Laser and Electronic Systems GmbH, Tübingen, Germany

HighFinesse Laser and Electronic Systems GmbH (hereinafter referred to as
“HighFinesse”) hereby grants, free of charge, anyone who downloads the Python dll
loading functions “wlmData.py” and “wlmConst.py” (hereinafter referred to as
“software”) from the HighFinesse website at https://www.highfinesse-downloads.com/download/t0849yd8uzpj or who receives it in any other way from
HighFinesse, a non-exclusive, non-transferable and non-sublicensable right, unlimited in
time and place, to use the software in any way or form, whether known or unknown, now
or in the future. This includes in particular the right to edit, modify, present, demonstrate,
reproduce, publish, use, digitize, distribute, broadcast, exhibit, make publicly accessible
and exploit the software.

These rights are granted only on condition that the following requirements are met:
• Any copy of the source code of the software must retain these licence terms in its
entirety. Furthermore, any edits, modifications and other changes in the copy
compared to software originally provided by HighFinesse must be clearly and
comprehensibly marked within the source code as “modified, deviating from
original files provided by HighFinesse”.
• Any copy of the software in binary form must reproduce these licence terms in its
entirety in the documentation and/or other materials provided with or connected
to the copy. Furthermore, any edits, modifications and other changes in the copy
compared to software originally provided by HighFinesse must be clearly and
comprehensibly marked as “modified, deviating from original files provided by
HighFinesse” within the documentation and/or other materials provided with or
connected to the copy.

The law of the Federal Republic of Germany shall apply exclusively. The provisions of the
UN Convention on Contracts for the International Sale of Goods (CISG) are excluded.

The courts of the Federal Republic of Germany shall have exclusive international
jurisdiction for all disputes arising from or in connection with these License terms. Among
the courts of the Federal Republic of Germany, the court having jurisdiction over the place
of business of HighFinesse shall have exclusive local jurisdiction for all disputes arising
from or in connection with these Licence terms.

HighFinesse would like to point out that the software is only a preliminary prototype and
not a final software product. HighFinesse is making this prototype available to the
experienced research community free of charge in order to promote research and enable
a critical discussion of the advantages, disadvantages and possible errors of the
software. The prototype is still under development at HighFinesse and has not yet
undergone any quality control. The software may therefore still contain errors, lack
essential functions or have other deficiencies. It should therefore only be used by highly
qualified software developers and should be extensively tested and checked before each
use in order to rule out errors and avoid any damage.

HighFinesse would be grateful for any information on errors and any suggestions for
improvement sent to info@highfinesse.de.
"""
#
# wlmData API constants generated from wlmData.h
#

# Instantiating Constants for 'RFC' parameter
cInstCheckForWLM = -1
cInstResetCalc = 0
cInstReturnMode = cInstResetCalc
cInstNotification = 1
cInstCopyPattern = 2
cInstCopyAnalysis = cInstCopyPattern
cInstControlWLM = 3
cInstControlDelay = 4
cInstControlPriority = 5

# Notification Constants for 'Mode' parameter
cNotifyInstallCallback = 0
cNotifyRemoveCallback = 1
cNotifyInstallWaitEvent = 2
cNotifyRemoveWaitEvent = 3
cNotifyInstallCallbackEx = 4
cNotifyInstallWaitEventEx = 5

# ResultError Constants of Set...-functions
ResERR_NoErr = 0
ResERR_WlmMissing = -1
ResERR_CouldNotSet = -2
ResERR_ParmOutOfRange = -3
ResERR_WlmOutOfResources = -4
ResERR_WlmInternalError = -5
ResERR_NotAvailable = -6
ResERR_WlmBusy = -7
ResERR_NotInMeasurementMode = -8
ResERR_OnlyInMeasurementMode = -9
ResERR_ChannelNotAvailable = -10
ResERR_ChannelTemporarilyNotAvailable = -11
ResERR_CalOptionNotAvailable = -12
ResERR_CalWavelengthOutOfRange = -13
ResERR_BadCalibrationSignal = -14
ResERR_UnitNotAvailable = -15
ResERR_FileNotFound = -16
ResERR_FileCreation = -17
ResERR_TriggerPending = -18
ResERR_TriggerWaiting = -19
ResERR_NoLegitimation = -20
ResERR_NoTCPLegitimation = -21
ResERR_NotInPulseMode = -22
ResERR_OnlyInPulseMode = -23
ResERR_NotInSwitchMode = -24
ResERR_OnlyInSwitchMode = -25
ResERR_TCPErr = -26

# Mode Constants for Callback-Export and WaitForWLMEvent-function
cmiResultMode = 1
cmiRange = 2
cmiPulse = 3
cmiPulseMode = cmiPulse
cmiWideLine = 4
cmiWideMode = cmiWideLine
cmiFast = 5
cmiFastMode = cmiFast
cmiExposureMode = 6
cmiExposureValue1 = 7
cmiExposureValue2 = 8
cmiDelay = 9
cmiShift = 10
cmiShift2 = 11
cmiReduce = 12
cmiReduced = cmiReduce
cmiScale = 13
cmiTemperature = 14
cmiLink = 15
cmiOperation = 16
cmiDisplayMode = 17
cmiPattern1a = 18
cmiPattern1b = 19
cmiPattern2a = 20
cmiPattern2b = 21
cmiMin1 = 22
cmiMax1 = 23
cmiMin2 = 24
cmiMax2 = 25
cmiNowTick = 26
cmiCallback = 27
cmiFrequency1 = 28
cmiFrequency2 = 29
cmiDLLDetach = 30
cmiVersion = 31
cmiAnalysisMode = 32
cmiDeviationMode = 33
cmiDeviationReference = 34
cmiDeviationSensitivity = 35
cmiAppearance = 36
cmiAutoCalMode = 37
cmiWavelength1 = 42
cmiWavelength2 = 43
cmiLinewidth = 44
cmiLinewidthMode = 45
cmiLinkDlg = 56
cmiAnalysis = 57
cmiAnalogIn = 66
cmiAnalogOut = 67
cmiDistance = 69
cmiWavelength3 = 90
cmiWavelength4 = 91
cmiWavelength5 = 92
cmiWavelength6 = 93
cmiWavelength7 = 94
cmiWavelength8 = 95
cmiVersion0 = cmiVersion
cmiVersion1 = 96
cmiPulseDelay = 99

cmiDLLAttach = 121
cmiSwitcherSignal = 123
cmiSwitcherMode = 124
cmiExposureValue11 = cmiExposureValue1
cmiExposureValue12 = 125
cmiExposureValue13 = 126
cmiExposureValue14 = 127
cmiExposureValue15 = 128
cmiExposureValue16 = 129
cmiExposureValue17 = 130
cmiExposureValue18 = 131
cmiExposureValue21 = cmiExposureValue2
cmiExposureValue22 = 132
cmiExposureValue23 = 133
cmiExposureValue24 = 134
cmiExposureValue25 = 135
cmiExposureValue26 = 136
cmiExposureValue27 = 137
cmiExposureValue28 = 138
cmiPatternAverage = 139
cmiPatternAvg1 = 140
cmiPatternAvg2 = 141
cmiAnalogOut1 = cmiAnalogOut
cmiAnalogOut2 = 142
cmiMin11 = cmiMin1
cmiMin12 = 146
cmiMin13 = 147
cmiMin14 = 148
cmiMin15 = 149
cmiMin16 = 150
cmiMin17 = 151
cmiMin18 = 152
cmiMin21 = cmiMin2
cmiMin22 = 153
cmiMin23 = 154
cmiMin24 = 155
cmiMin25 = 156
cmiMin26 = 157
cmiMin27 = 158
cmiMin28 = 159
cmiMax11 = cmiMax1
cmiMax12 = 160
cmiMax13 = 161
cmiMax14 = 162
cmiMax15 = 163
cmiMax16 = 164
cmiMax17 = 165
cmiMax18 = 166
cmiMax21 = cmiMax2
cmiMax22 = 167
cmiMax23 = 168
cmiMax24 = 169
cmiMax25 = 170
cmiMax26 = 171
cmiMax27 = 172
cmiMax28 = 173
cmiAvg11 = cmiPatternAvg1
cmiAvg12 = 174
cmiAvg13 = 175
cmiAvg14 = 176
cmiAvg15 = 177
cmiAvg16 = 178
cmiAvg17 = 179
cmiAvg18 = 180
cmiAvg21 = cmiPatternAvg2
cmiAvg22 = 181
cmiAvg23 = 182
cmiAvg24 = 183
cmiAvg25 = 184
cmiAvg26 = 185
cmiAvg27 = 186
cmiAvg28 = 187
cmiPatternAnalysisWritten = 202
cmiSwitcherChannel = 203
cmiStartCalibration = 235
cmiEndCalibration = 236
cmiAnalogOut3 = 237
cmiAnalogOut4 = 238
cmiAnalogOut5 = 239
cmiAnalogOut6 = 240
cmiAnalogOut7 = 241
cmiAnalogOut8 = 242
cmiIntensity = 251
cmiPower = 267
cmiActiveChannel = 300
cmiPIDCourse = 1030
cmiPIDUseTa = 1031
cmiPIDUseT = cmiPIDUseTa
cmiPID_T = 1033
cmiPID_P = 1034
cmiPID_I = 1035
cmiPID_D = 1036
cmiDeviationSensitivityDim = 1040
cmiDeviationSensitivityFactor = 1037
cmiDeviationPolarity = 1038
cmiDeviationSensitivityEx = 1039
cmiDeviationUnit = 1041
cmiDeviationBoundsMin = 1042
cmiDeviationBoundsMax = 1043
cmiDeviationRefMid = 1044
cmiDeviationRefAt = 1045
cmiPIDConstdt = 1059
cmiPID_dt = 1060
cmiPID_AutoClearHistory = 1061
cmiDeviationChannel = 1063
cmiPID_ClearHistoryOnRangeExceed = 1069
cmiAutoCalPeriod = 1120
cmiAutoCalUnit = 1121
cmiAutoCalChannel = 1122
cmiServerInitialized = 1124
cmiWavelength9 = 1130
cmiExposureValue19 = 1155
cmiExposureValue29 = 1180
cmiMin19 = 1205
cmiMin29 = 1230
cmiMax19 = 1255
cmiMax29 = 1280
cmiAvg19 = 1305
cmiAvg29 = 1330
cmiWavelength10 = 1355
cmiWavelength11 = 1356
cmiWavelength12 = 1357
cmiWavelength13 = 1358
cmiWavelength14 = 1359
cmiWavelength15 = 1360
cmiWavelength16 = 1361
cmiWavelength17 = 1362
cmiExternalInput = 1400
cmiPressure = 1465
cmiBackground = 1475
cmiDistanceMode = 1476
cmiInterval = 1477
cmiIntervalMode = 1478
cmiCalibrationEffect = 1480
cmiLinewidth1 = cmiLinewidth
cmiLinewidth2 = 1481
cmiLinewidth3 = 1482
cmiLinewidth4 = 1483
cmiLinewidth5 = 1484
cmiLinewidth6 = 1485
cmiLinewidth7 = 1486
cmiLinewidth8 = 1487
cmiLinewidth9 = 1488
cmiLinewidth10 = 1489
cmiLinewidth11 = 1490
cmiLinewidth12 = 1491
cmiLinewidth13 = 1492
cmiLinewidth14 = 1493
cmiLinewidth15 = 1494
cmiLinewidth16 = 1495
cmiLinewidth17 = 1496
cmiTriggerState = 1497
cmiDeviceAttach = 1501
cmiDeviceDetach = 1502
cmiTimePerMeasurement = 1514
cmiAutoExpoMin = 1517
cmiAutoExpoMax = 1518
cmiAutoExpoStepUp = 1519
cmiAutoExpoStepDown = 1520
cmiAutoExpoAtSaturation = 1521
cmiAutoExpoAtLowSignal = 1522
cmiAutoExpoFeedback = 1523
cmiAveragingCount = 1524
cmiAveragingMode = 1525
cmiAveragingType = 1526
cmiNowTick_d  = 1527
cmiAirMode = 1532
cmiAirTemperature = 1534
cmiAirPressure = 1535
cmiAirHumidity = 1536
cmiAirCO2 = 1651
cmiSubSnapshotID = 1539
cmiInternalTriggerRate = 1540
cmiGain11 = 1541
cmiGain12 = 1542
cmiGain13 = 1543
cmiGain14 = 1544
cmiGain15 = 1545
cmiGain16 = 1546
cmiGain17 = 1547
cmiGain18 = 1548
cmiGain19 = 1549
cmiGain110 = 1550
cmiGain111 = 1551
cmiGain112 = 1552
cmiGain113 = 1553
cmiGain114 = 1554
cmiGain115 = 1555
cmiGain116 = 1556
cmiGain117 = 1557
cmiGain21 = 1558
cmiGain22 = 1559
cmiGain23 = 1560
cmiGain24 = 1561
cmiGain25 = 1562
cmiGain26 = 1563
cmiGain27 = 1564
cmiGain28 = 1565
cmiGain29 = 1566
cmiGain210 = 1567
cmiGain211 = 1568
cmiGain212 = 1569
cmiGain213 = 1570
cmiGain214 = 1571
cmiGain215 = 1572
cmiGain216 = 1573
cmiGain217 = 1574
cmiGain31 = 1575
cmiGain32 = 1576
cmiGain33 = 1577
cmiGain34 = 1578
cmiGain35 = 1579
cmiGain36 = 1580
cmiGain37 = 1581
cmiGain38 = 1582
cmiGain39 = 1583
cmiGain310 = 1584
cmiGain311 = 1585
cmiGain312 = 1586
cmiGain313 = 1587
cmiGain314 = 1588
cmiGain315 = 1589
cmiGain316 = 1590
cmiGain317 = 1591
cmiGain41 = 1592
cmiGain42 = 1593
cmiGain43 = 1594
cmiGain44 = 1595
cmiGain45 = 1596
cmiGain46 = 1597
cmiGain47 = 1598
cmiGain48 = 1599
cmiGain49 = 1600
cmiGain410 = 1601
cmiGain411 = 1602
cmiGain412 = 1603
cmiGain413 = 1604
cmiGain414 = 1605
cmiGain415 = 1606
cmiGain416 = 1607
cmiGain417 = 1608
cmiMultimodeLevel1 = 1609
cmiMultimodeLevel2 = 1610
cmiMultimodeLevel3 = 1611
cmiMultimodeLevel4 = 1612
cmiMultimodeLevel5 = 1613
cmiMultimodeLevel6 = 1614
cmiMultimodeLevel7 = 1615
cmiMultimodeLevel8 = 1616
cmiMultimodeLevel9 = 1617
cmiMultimodeLevel10 = 1618
cmiMultimodeLevel11 = 1619
cmiMultimodeLevel12 = 1620
cmiMultimodeLevel13 = 1621
cmiMultimodeLevel14 = 1622
cmiMultimodeLevel15 = 1623
cmiMultimodeLevel16 = 1624
cmiMultimodeLevel17 = 1625
cmiFastBasedLinewidthAnalysis = 1630
cmiMultimodeCount1 = 1633
cmiMultimodeCount2 = 1634
cmiMultimodeCount3 = 1635
cmiMultimodeCount4 = 1636
cmiMultimodeCount5 = 1637
cmiMultimodeCount6 = 1638
cmiMultimodeCount7 = 1639
cmiMultimodeCount8 = 1640
cmiMultimodeCount9 = 1641
cmiMultimodeCount10 = 1642
cmiMultimodeCount11 = 1643
cmiMultimodeCount12 = 1644
cmiMultimodeCount13 = 1645
cmiMultimodeCount14 = 1646
cmiMultimodeCount15 = 1647
cmiMultimodeCount16 = 1648
cmiMultimodeCount17 = 1649

# Index constants for Get- and SetExtraSetting
cesCalculateLive = 4501

# WLM Control Mode Constants
cCtrlWLMShow = 1
cCtrlWLMHide = 2
cCtrlWLMExit = 3
cCtrlWLMStore = 4
cCtrlWLMCompare = 5
cCtrlWLMWait        = 0x0010
cCtrlWLMStartSilent = 0x0020
cCtrlWLMSilent      = 0x0040
cCtrlWLMStartDelay  = 0x0080

# Operation Mode Constants (for "Operation" and "GetOperationState" functions)
cStop = 0
cAdjustment = 1
cMeasurement = 2

# Base Operation Constants (To be used exclusively, only one of this list at a time,
# but still can be combined with "Measurement Action Addition Constants". See below.)
cCtrlStopAll = cStop
cCtrlStartAdjustment = cAdjustment
cCtrlStartMeasurement = cMeasurement
cCtrlStartRecord = 0x0004
cCtrlStartReplay = 0x0008
cCtrlStoreArray  = 0x0010
cCtrlLoadArray   = 0x0020

# Additional Operation Flag Constants (combine with "Base Operation Constants" above.)
cCtrlDontOverwrite = 0x0000

cCtrlFileGiven     = 0x0000


# Measurement Control Mode Constants
cCtrlMeasDelayRemove = 0
cCtrlMeasDelayGenerally = 1
cCtrlMeasDelayOnce = 2
cCtrlMeasDelayDenyUntil = 3
cCtrlMeasDelayIdleOnce = 4
cCtrlMeasDelayIdleEach = 5
cCtrlMeasDelayDefault = 6

# Measurement Triggering Action Constants
cCtrlMeasurementContinue = 0
cCtrlMeasurementInterrupt = 1
cCtrlMeasurementTriggerPoll = 2
cCtrlMeasurementTriggerSuccess = 3
cCtrlMeasurementEx = 0x0100

# ExposureRange Constants
cExpoMin = 0
cExpoMax = 1
cExpo2Min = 2
cExpo2Max = 3

# Amplitude Constants
cMin1 = 0
cMin2 = 1
cMax1 = 2
cMax2 = 3
cAvg1 = 4
cAvg2 = 5

# Measurement Range Constants
cRange_250_410 = 4
cRange_250_425 = 0
cRange_300_410 = 3
cRange_350_500 = 5
cRange_400_725 = 1
cRange_700_1100 = 2
cRange_800_1300 = 6
cRange_900_1500 = cRange_800_1300
cRange_1100_1700 = 7
cRange_1100_1800 = cRange_1100_1700

# Measurement Range Model Constants
cRangeModelOld = 65535
cRangeModelByOrder = 65534
cRangeModelByWavelength = 65533

# Unit Constants for Get-/SetResultMode, GetLinewidth, Convert... and Calibration
cReturnWavelengthVac = 0
cReturnWavelengthAir = 1
cReturnFrequency = 2
cReturnWavenumber = 3
cReturnPhotonEnergy = 4

# Power Unit Constants
cPower_muW = 0
cPower_dBm = 1

# Source Type Constants for Calibration
cHeNe633 = 0
cHeNe1152 = 0
cNeL = 1
cOther = 2
cFreeHeNe = 3
cSLR1530 = 5

# Unit Constants for Autocalibration
cACOnceOnStart = 0
cACMeasurements = 1
cACDays = 2
cACHours = 3
cACMinutes = 4

# ExposureRange Constants
cGetSync = 1
cSetSync = 2

# Pattern- and Analysis Constants
cPatternDisable = 0
cPatternEnable = 1
cAnalysisDisable = cPatternDisable
cAnalysisEnable = cPatternEnable

cSignal1Interferometers = 0
cSignal1WideInterferometer = 1
cSignal1Grating = 1
cSignal2Interferometers = 2
cSignal2WideInterferometer = 3
cSignalAnalysis = 4
cSignalAnalysisX = cSignalAnalysis
cSignalAnalysisY = cSignalAnalysis + 1

# State constants used with AutoExposureSetting functions
cJustStepDown = 0
cRestartAtMinimum = 1
cJustStepUp = 0
cDriveToLevel = 1
cConsiderFeedback = 1
cDontConsiderFeedback = 0

# Options identifiers used with GetOptionInfo
cInfoSwitch = 1
cInfoSwitchChannelsCount = 2
cInfoIntNeonLamp = 11
cInfo2ndExternalPort = 13
cInfoPID = 21
cInfoPIDPortsCount = 22
cInfoPIDPortType = 23
cInfoPIDPortRes = 24
cInfoPIDPortUMin = 25
cInfoPIDPortUMax = 26

# PID type constants
cInfoPIDPortTypeInt = 1
cInfoPIDPortTypeExt = 2
cInfoPIDPortTypeDigi = 3

# State constants used with AveragingSetting functions
cAvrgFloating = 1
cAvrgSucceeding = 2
cAvrgSimple = 0
cAvrgPattern = 1

# Return errorvalues of GetFrequency, GetWavelength, GetWLMVersion and GetOptionInfo
ErrNoValue = 0
ErrNoSignal = -1
ErrBadSignal = -2
ErrLowSignal = -3
ErrBigSignal = -4
ErrWlmMissing = -5
ErrNotAvailable = -6
InfNothingChanged = -7
ErrNoPulse = -8
ErrChannelNotAvailable = -10
ErrDiv0 = -13
ErrOutOfRange = -14
ErrUnitNotAvailable = -15
ErrTCPErr = -26
ErrParameterOutOfRange = -28
ErrMaxErr = ErrParameterOutOfRange

# Return errorvalues of GetTemperature and GetPressure
ErrTemperature = -1000
ErrTempNotMeasured = ErrTemperature + ErrNoValue
ErrTempNotAvailable = ErrTemperature + ErrNotAvailable
ErrTempWlmMissing = ErrTemperature + ErrWlmMissing

# Return errorvalues of GetGain
ErrGain = -1000
ErrGainNotAvailable        = ErrGain + ErrNotAvailable
ErrGainWlmMissing          = ErrGain + ErrWlmMissing
ErrGainChannelNotAvailable = ErrGain + ErrChannelNotAvailable
ErrGainOutOfRange          = ErrGain + ErrOutOfRange
ErrGainParameterOutOfRange = ErrGain + ErrParameterOutOfRange

# Return errorvalues of GetMultimodeInfo
ErrMMI = -1000
ErrMMINotAvailable        = ErrMMI + ErrNotAvailable
ErrMMIWlmMissing          = ErrMMI + ErrWlmMissing
ErrMMIChannelNotAvailable = ErrMMI + ErrChannelNotAvailable
ErrMMIOutOfRange          = ErrMMI + ErrOutOfRange
ErrMMIParameterOutOfRange = ErrMMI + ErrParameterOutOfRange

# Return errorvalues of GetDistance
# real errorvalues are ErrDistance combined with those of GetWavelength
ErrDistance = -1000000000
ErrDistanceNotAvailable = ErrDistance + ErrNotAvailable
ErrDistanceWlmMissing = ErrDistance + ErrWlmMissing

# Return flags of ControlWLMEx in combination with Show or Hide, Wait and Res = 1
flServerStarted           = 0x00000001
flErrDeviceNotFound       = 0x00000002
flErrDriverError          = 0x00000004
flErrUSBError             = 0x00000008
flErrUnknownDeviceError   = 0x00000010
flErrWrongSN              = 0x00000020
flErrUnknownSN            = 0x00000040
flErrTemperatureError     = 0x00000080
flErrPressureError        = 0x00000100
flErrCancelledManually    = 0x00000200
flErrWLMBusy              = 0x00000400
flErrUnknownError         = 0x00001000
flNoInstalledVersionFound = 0x00002000
flDesiredVersionNotFound  = 0x00004000
flErrFileNotFound         = 0x00008000
flErrParmOutOfRange       = 0x00010000
flErrCouldNotSet          = 0x00020000
flErrEEPROMFailed         = 0x00040000
flErrFileFailed           = 0x00080000
flDeviceDataNewer         = 0x00100000
flFileDataNewer           = 0x00200000
flErrDeviceVersionOld     = 0x00400000
flErrFileVersionOld       = 0x00800000
flDeviceStampNewer        = 0x01000000
flFileStampNewer          = 0x02000000

# Return file info flags of SetOperationFile
flFileInfoDoesntExist = 0x0000
flFileInfoExists      = 0x0001
flFileInfoCantWrite   = 0x0002
flFileInfoCantRead    = 0x0004
flFileInfoInvalidName = 0x0008
cFileParameterError = -1


